"""
Tests all of dymos' docstrings using the Numpydocs validator.
"""
import ast
import importlib
import inspect
import os
import re
import sys
import unittest

try:
    from numpydoc.docscrape import NumpyDocString
    from numpydoc import validate
except ImportError:
    NumpyDocString = None


# Directories to exclude from docstring linting.
exclude = [
    'doc',
    'test',
    'examples',
    'plots',
    '_html',
    '__pycache__',
    'coloring_files',
]

# Error Codes to Ignore
ignore = [
    'ES01',    # No extended summary found
    'EX01',    # No examples section found
    'SA01',    # See Also section not found
    'SS05',    # Summary must start with infinitive verb, not third person (e.g. use "Generate" instead of "Generates")')
]

# Build a list of dirs in which to do linting.
directories = []

top = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

for root, dirs, files in os.walk(top, topdown=True):
    dirs[:] = [d for d in dirs if d not in exclude]
    for item in dirs:
        directories.append(os.path.join(root, item))


@unittest.skipUnless(NumpyDocString, "requires 'numpydoc' >= 1.1")
class DocstringLintTestCase(unittest.TestCase):

    def test_docstrings(self):
        failures = {}

        # Loop over directories
        for dirpath in sorted(directories):

            # Loop over files
            for file_name in sorted(os.listdir(dirpath)):
                if not file_name.startswith("_") and file_name[-3:] == '.py' \
                   and not os.path.isdir(file_name):

                    # To construct module name, remove part of abs path that
                    # precedes 'dymos', and then replace '/' with '.' in the remainder.
                    mod1 = re.sub(r'.*dymos', 'dymos', dirpath).replace('/', '.')

                    # Then, get rid of the '.py' to get final part of module name.
                    mod2 = file_name[:-3]

                    module_name = f'{mod1}.{mod2}'

                    try:
                        mod = importlib.import_module(module_name)
                    except ImportError:
                        # e.g. PETSc is not installed
                        failures[module_name] = [('EEE', f'Error when loading module {module_name}.')]
                        continue

                    classes = [x for x in dir(mod)
                               if not x.startswith('_') and inspect.isclass(getattr(mod, x)) and
                               getattr(mod, x).__module__ == module_name]

                    # Loop over classes.
                    for class_name in classes:
                        full_class_path = f'{module_name}.{class_name}'
                        try:
                            result = validate.validate(full_class_path)
                        except Exception:
                            continue

                        for error_tuple in result['errors']:
                            if error_tuple[0] not in ignore:
                                if full_class_path not in failures:
                                    failures[full_class_path] = []
                                msg = f"{error_tuple[0]}: {error_tuple[1]}"
                                failures[full_class_path].append(msg)

                        clss = getattr(mod, class_name)

                        methods = [x for x in dir(clss)
                                   if (inspect.ismethod(getattr(clss, x)) or
                                       inspect.isfunction(getattr(clss, x))) and
                                   x in clss.__dict__]

                        # Loop over class methods.
                        for method_name in methods:

                            if not method_name.startswith('_'):
                                full_method_path = f'{module_name}.{class_name}.{method_name}'
                                try:
                                    result = validate.validate(full_method_path)
                                except Exception:
                                    continue

                                for error_tuple in result['errors']:
                                    if error_tuple[0] not in ignore:
                                        if full_method_path not in failures:
                                            failures[full_method_path] = []
                                        msg = f"{error_tuple[0]}: {error_tuple[1]}"
                                        failures[full_method_path].append(msg)

                    try:
                        tree = ast.parse(inspect.getsource(mod))
                    except:
                        print(f'Error reading file {module_name}', file=sys.stderr)
                        raise

                    if hasattr(tree, 'body'):
                        funcs = [node.name for node in tree.body if isinstance(node, ast.FunctionDef)]
                    else:
                        funcs = []

                    # Loop over standalone functions.
                    for func_name in funcs:
                        if not func_name.startswith('_'):
                            full_function_path = f'{module_name}.{func_name}'
                            try:
                                result = validate.validate(full_function_path)
                            except Exception:
                                continue

                            for error_tuple in result['errors']:
                                if error_tuple[0] not in ignore:
                                    if full_function_path not in failures:
                                        failures[full_function_path] = []
                                    msg = f"{error_tuple[0]}: {error_tuple[1]}"
                                    failures[full_function_path].append(msg)

        if failures:
            msg = '\n'
            count = 0
            for key in failures:
                msg += f'{key}\n'
                count += len(failures[key])
                for failure in failures[key]:
                    msg += f'    {failure}\n'
            msg += f'Found {count} issues in docstrings'
            self.fail(msg)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import openmdao.api as om


class HullProblemODE(om.ExplicitComponent):
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input('u', val=np.zeros(nn), desc='control')

        self.add_output('L', val=np.zeros(nn), desc='Lagrangian', units='1/s')

        # Setup partials
        self.declare_partials(of='L', wrt='u', rows=np.arange(nn), cols=np.arange(nn))

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        u = inputs['u']

        outputs['L'] = 0.5 * u ** 2

    def compute_partials(self, inputs, partials, discrete_inputs=None):
        u = inputs['u']
        partials['L', 'u'] = u



import unittest

from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse
from openmdao.utils.assert_utils import assert_near_equal

from dymos.utils.misc import om_version


SHOW_PLOTS = True


@use_tempdirs
class TestBalancedFieldLengthForDocs(unittest.TestCase):

    @require_pyoptsparse(optimizer='IPOPT')
    def test_balanced_field_length_for_docs(self):
        import openmdao.api as om
        from openmdao.utils.general_utils import set_pyoptsparse_opt
        import dymos as dm
        from dymos.examples.balanced_field.balanced_field_ode import BalancedFieldODEComp

        p = om.Problem()

        _, optimizer = set_pyoptsparse_opt('IPOPT', fallback=True)

        dm.options['plots'] = 'bokeh'

        # Use IPOPT if available, with fallback to SLSQP
        p.driver = om.pyOptSparseDriver(optimizer='SNOPT')
        p.driver.options['optimizer'] = optimizer
        p.driver.declare_coloring()
        p.driver.options['print_results'] = False
        if optimizer == 'IPOPT':
            p.driver.opt_settings['print_level'] = 0
            p.driver.opt_settings['derivative_test'] = 'first-order'

        # First Phase: Brake release to V1 - both engines operable
        br_to_v1 = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=3),
                            ode_init_kwargs={'mode': 'runway'})
        br_to_v1.set_time_options(fix_initial=True, duration_bounds=(1, 1000), duration_ref=10.0)
        br_to_v1.add_state('r', fix_initial=True, lower=0, ref=1000.0, defect_ref=1000.0)
        br_to_v1.add_state('v', fix_initial=True, lower=0, ref=100.0, defect_ref=100.0)
        br_to_v1.add_parameter('alpha', val=0.0, opt=False, units='deg')
        br_to_v1.add_timeseries_output('*')

        # Second Phase: Rejected takeoff at V1 - no engines operable
        rto = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=3),
                       ode_init_kwargs={'mode': 'runway'})
        rto.set_time_options(fix_initial=False, duration_bounds=(1, 1000), duration_ref=1.0)
        rto.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        rto.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
        rto.add_parameter('alpha', val=0.0, opt=False, units='deg')
        rto.add_timeseries_output('*')

        # Third Phase: V1 to Vr - single engine operable
        v1_to_vr = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=3),
                            ode_init_kwargs={'mode': 'runway'})
        v1_to_vr.set_time_options(fix_initial=False, duration_bounds=(1, 1000), duration_ref=1.0)
        v1_to_vr.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        v1_to_vr.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
        v1_to_vr.add_parameter('alpha', val=0.0, opt=False, units='deg')
        v1_to_vr.add_timeseries_output('*')

        # Fourth Phase: Rotate - single engine operable
        rotate = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=3),
                          ode_init_kwargs={'mode': 'runway'})
        rotate.set_time_options(fix_initial=False, duration_bounds=(1.0, 5), duration_ref=1.0)
        rotate.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        rotate.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
        rotate.add_control('alpha', order=1, opt=True, units='deg', lower=0, upper=10, ref=10, val=[0, 10],
                           control_type='polynomial')
        rotate.add_timeseries_output('*')

        # Fifth Phase: Climb to target speed and altitude at end of runway.
        climb = dm.Phase(ode_class=BalancedFieldODEComp, transcription=dm.Radau(num_segments=5),
                         ode_init_kwargs={'mode': 'climb'})
        climb.set_time_options(fix_initial=False, duration_bounds=(1, 100), duration_ref=1.0)
        climb.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        climb.add_state('h', fix_initial=True, lower=0, ref=1.0, defect_ref=1.0)
        climb.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
        climb.add_state('gam', fix_initial=True, lower=0, ref=0.05, defect_ref=0.05)
        climb.add_control('alpha', opt=True, units='deg', lower=-10, upper=15, ref=10)
        climb.add_timeseries_output('*')

        # Instantiate the trajectory and add phases
        traj = dm.Trajectory()
        p.model.add_subsystem('traj', traj)
        traj.add_phase('br_to_v1', br_to_v1)
        traj.add_phase('rto', rto)
        traj.add_phase('v1_to_vr', v1_to_vr)
        traj.add_phase('rotate', rotate)
        traj.add_phase('climb', climb)

        # Add parameters common to multiple phases to the trajectory
        traj.add_parameter('m', val=174200., opt=False, units='lbm',
                           desc='aircraft mass',
                           targets={'br_to_v1': ['m'], 'v1_to_vr': ['m'], 'rto': ['m'],
                                    'rotate': ['m'], 'climb': ['m']})

        traj.add_parameter('T_nominal', val=27000 * 2, opt=False, units='lbf', static_target=True,
                           desc='nominal aircraft thrust',
                           targets={'br_to_v1': ['T']})

        traj.add_parameter('T_engine_out', val=27000, opt=False, units='lbf', static_target=True,
                           desc='thrust under a single engine',
                           targets={'v1_to_vr': ['T'], 'rotate': ['T'], 'climb': ['T']})

        traj.add_parameter('T_shutdown', val=0.0, opt=False, units='lbf', static_target=True,
                           desc='thrust when engines are shut down for rejected takeoff',
                           targets={'rto': ['T']})

        traj.add_parameter('mu_r_nominal', val=0.03, opt=False, units=None, static_target=True,
                           desc='nominal runway friction coefficient',
                           targets={'br_to_v1': ['mu_r'], 'v1_to_vr': ['mu_r'],  'rotate': ['mu_r']})

        traj.add_parameter('mu_r_braking', val=0.3, opt=False, units=None, static_target=True,
                           desc='runway friction coefficient under braking',
                           targets={'rto': ['mu_r']})

        traj.add_parameter('h_runway', val=0., opt=False, units='ft',
                           desc='runway altitude',
                           targets={'br_to_v1': ['h'], 'v1_to_vr': ['h'], 'rto': ['h'],
                                    'rotate': ['h']})

        traj.add_parameter('rho', val=1.225, opt=False, units='kg/m**3', static_target=True,
                           desc='atmospheric density',
                           targets={'br_to_v1': ['rho'], 'v1_to_vr': ['rho'], 'rto': ['rho'],
                                    'rotate': ['rho']})

        traj.add_parameter('S', val=124.7, opt=False, units='m**2', static_target=True,
                           desc='aerodynamic reference area',
                           targets={'br_to_v1': ['S'], 'v1_to_vr': ['S'], 'rto': ['S'],
                                    'rotate': ['S'], 'climb': ['S']})

        traj.add_parameter('CD0', val=0.03, opt=False, units=None, static_target=True,
                           desc='zero-lift drag coefficient',
                           targets={f'{phase}': ['CD0'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                      'rto', 'rotate', 'climb']})

        traj.add_parameter('AR', val=9.45, opt=False, units=None, static_target=True,
                           desc='wing aspect ratio',
                           targets={f'{phase}': ['AR'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                     'rto', 'rotate', 'climb']})

        traj.add_parameter('e', val=801, opt=False, units=None, static_target=True,
                           desc='Oswald span efficiency factor',
                           targets={f'{phase}': ['e'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                    'rto', 'rotate', 'climb']})

        traj.add_parameter('span', val=35.7, opt=False, units='m', static_target=True,
                           desc='wingspan',
                           targets={f'{phase}': ['span'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                       'rto', 'rotate', 'climb']})

        traj.add_parameter('h_w', val=1.0, opt=False, units='m', static_target=True,
                           desc='height of wing above CG',
                           targets={f'{phase}': ['h_w'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                      'rto', 'rotate', 'climb']})

        traj.add_parameter('CL0', val=0.5, opt=False, units=None, static_target=True,
                           desc='zero-alpha lift coefficient',
                           targets={f'{phase}': ['CL0'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                      'rto', 'rotate', 'climb']})

        traj.add_parameter('CL_max', val=2.0, opt=False, units=None, static_target=True,
                           desc='maximum lift coefficient for linear fit',
                           targets={f'{phase}': ['CL_max'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                         'rto', 'rotate', 'climb']})

        traj.add_parameter('alpha_max', val=10.0, opt=False, units='deg', static_target=True,
                           desc='angle of attack at maximum lift',
                           targets={f'{phase}': ['alpha_max'] for phase in ['br_to_v1', 'v1_to_vr',
                                                                            'rto', 'rotate', 'climb']})

        # Standard "end of first phase to beginning of second phase" linkages
        # Alpha changes from being a parameter in v1_to_vr to a polynomial control
        # in rotate, to a dynamic control in `climb`.
        traj.link_phases(['br_to_v1', 'v1_to_vr'], vars=['time', 'r', 'v'])
        traj.link_phases(['v1_to_vr', 'rotate'], vars=['time', 'r', 'v', 'alpha'])
        traj.link_phases(['rotate', 'climb'], vars=['time', 'r', 'v', 'alpha'])
        traj.link_phases(['br_to_v1', 'rto'], vars=['time', 'r', 'v'])

        # Less common "final value of r must be the match at ends of two phases".
        traj.add_linkage_constraint(phase_a='rto', var_a='r', loc_a='final',
                                    phase_b='climb', var_b='r', loc_b='final',
                                    ref=1000)

        # Define the constraints and objective for the optimal control problem
        v1_to_vr.add_boundary_constraint('v_over_v_stall', loc='final', lower=1.2, ref=100)

        rto.add_boundary_constraint('v', loc='final', equals=0., ref=100, linear=True)

        rotate.add_boundary_constraint('F_r', loc='final', equals=0, ref=100000)

        climb.add_boundary_constraint('h', loc='final', equals=35, ref=35, units='ft', linear=True)
        climb.add_boundary_constraint('gam', loc='final', equals=5, ref=5, units='deg', linear=True)
        climb.add_path_constraint('gam', lower=0, upper=5, ref=5, units='deg')
        climb.add_boundary_constraint('v_over_v_stall', loc='final', lower=1.25, ref=1.25)

        rto.add_objective('r', loc='final', ref=1.0)

        for phase_name, phase in traj._phases.items():
            if 'T_nominal' in phase.parameter_options:
                phase.add_timeseries_output('T_nominal', output_name='T')
            if 'T_engine_out' in phase.parameter_options:
                phase.add_timeseries_output('T_engine_out', output_name='T')
            if 'T_shutdown' in phase.parameter_options:
                phase.add_timeseries_output('T_shutdown', output_name='T')
            phase.add_timeseries_output('alpha')

        #
        # Setup the problem and set the initial guess
        #
        p.setup(check=True)

        br_to_v1.set_time_val(initial=0.0, duration=35.0)
        br_to_v1.set_state_val('r', [0, 2500.0])
        br_to_v1.set_state_val('v', [0.0, 100.0])
        br_to_v1.set_parameter_val('alpha', 0.0, units='deg')

        v1_to_vr.set_time_val(initial=35.0, duration=35.0)
        v1_to_vr.set_state_val('r', [2500, 300.0])
        v1_to_vr.set_state_val('v', [100, 110.0])
        v1_to_vr.set_parameter_val('alpha', 0.0, units='deg')

        rto.set_time_val(initial=35.0, duration=35.0)
        rto.set_state_val('r', [2500, 5000.0])
        rto.set_state_val('v', [110, 0.0])
        rto.set_parameter_val('alpha', 0.0, units='deg')

        rotate.set_time_val(initial=70.0, duration=5.0)
        rotate.set_state_val('r', [1750, 1800.0])
        rotate.set_state_val('v', [80, 85.0])
        rotate.set_control_val('alpha', 0.0, units='deg')

        climb.set_time_val(initial=75.0, duration=15.0)
        climb.set_state_val('r', [5000, 5500.0], units='ft')
        climb.set_state_val('v', [160, 170.0], units='kn')
        climb.set_state_val('h', [0.0, 35.0], units='ft')
        climb.set_state_val('gam', [0.0, 5.0], units='deg')
        climb.set_control_val('alpha', 5.0, units='deg')

        dm.run_problem(p, run_driver=True, simulate=True, make_plots=True)

        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = traj.sim_prob.get_outputs_dir() / sim_db

        sol = om.CaseReader(sol_db).get_case('final')
        sim = om.CaseReader(sim_db).get_case('final')

        sol_r_f_climb = sol.get_val('traj.climb.timeseries.r')[-1, ...]
        sol_r_f_rto = sol.get_val('traj.rto.timeseries.r')[-1, ...]
        sim_r_f_climb = sim.get_val('traj.climb.timeseries.r')[-1, ...]
        sim_r_f_rto = sim.get_val('traj.rto.timeseries.r')[-1, ...]

        assert_near_equal(2114.387, sol_r_f_climb, tolerance=0.01)
        assert_near_equal(2114.387, sol_r_f_rto, tolerance=0.01)
        assert_near_equal(2114.387, sim_r_f_climb, tolerance=0.01)
        assert_near_equal(2114.387, sim_r_f_rto, tolerance=0.01)



import unittest

import numpy as np
import openmdao.api as om
import dymos as dm

from openmdao.utils.testing_utils import use_tempdirs
from openmdao.utils.assert_utils import assert_warning


def make_phases(traj, time_option_kwargs):
    for pname, kwargs in time_option_kwargs.items():
        phase = dm.Phase(ode_class=OscillatorODE, transcription=dm.Radau(num_segments=10))
        traj.add_phase(pname, phase)

        phase.set_time_options(**kwargs)

        # Tell Dymos the states to be propagated using the given ODE.
        phase.add_state('v', rate_source='v_dot', targets=['v'], units='m/s')
        phase.add_state('x', rate_source='v', targets=['x'], units='m')

        # The spring constant, damping coefficient, and mass are inputs to the system
        # that are constant throughout the phase.
        phase.add_parameter('k', units='N/m', targets=['k'])
        phase.add_parameter('c', units='N*s/m', targets=['c'])
        phase.add_parameter('m', units='kg', targets=['m'])

    # Since we're using an optimization driver, an objective is required.  We'll minimize
    # the final time in this case.
    phase.add_objective('time', loc='final')


def connect_phases(traj, conns):
    for src, tgts in conns:
        for tgt in tgts:
            traj.link_phases((src, tgt), vars=['time'])


class OscillatorODE(om.ExplicitComponent):
    """
    A Dymos ODE for a damped harmonic oscillator.
    """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input('x', shape=(nn,), desc='displacement', units='m')
        self.add_input('v', shape=(nn,), desc='velocity', units='m/s')
        self.add_input('k', shape=(nn,), desc='spring constant', units='N/m')
        self.add_input('c', shape=(nn,), desc='damping coefficient', units='N*s/m')
        self.add_input('m', shape=(nn,), desc='mass', units='kg')

        self.add_output('v_dot', val=np.ones(nn), desc='rate of change of velocity', units='m/s**2')

        arange = np.arange(self.options['num_nodes'], dtype=int)
        self.declare_partials(of='*', wrt='*', rows=arange, cols=arange)

    def compute(self, inputs, outputs):
        x = inputs['x']
        v = inputs['v']
        k = inputs['k']
        c = inputs['c']
        m = inputs['m']

        f_spring = -k * x
        f_damper = -c * v

        outputs['v_dot'] = (f_spring + f_damper) / m

    def compute_partials(self, inputs, jacobian):
        x = inputs['x']
        v = inputs['v']
        k = inputs['k']
        c = inputs['c']
        m = inputs['m']

        jacobian['v_dot', 'x'] = -k / m
        jacobian['v_dot', 'v'] = -c / m
        jacobian['v_dot', 'k'] = -x / m
        jacobian['v_dot', 'c'] = -v / m
        jacobian['v_dot', 'm'] = -1. / (m * m) * (-k * x - c * v)


@use_tempdirs
class Test_t_initialBounds(unittest.TestCase):
    def try_model(self, probname, kwargs, conns):
        prob = om.Problem(name=probname)
        prob.driver = om.ScipyOptimizeDriver()
        traj = prob.model.add_subsystem('traj', dm.Trajectory())

        make_phases(traj, kwargs)
        connect_phases(traj, conns)

        prob.setup(force_alloc_complex=True)

        dm.run_problem(prob, run_driver=False)

    def test_pair_fixed_t_initial_below(self):
        kwargs = {
            'phase0': {
                'fix_initial': True,
                'initial_val': 5.,
                'initial_bounds': (None, None),
                'fix_duration': False,
                'duration_bounds': (5., 10.),
            },
            'phase1': {
                'fix_initial': True,
                'initial_val': 5.,
                'initial_bounds': (None, None),
                'fix_duration': False,
                'duration_bounds': (5., 10.),
            }
        }
        conns = [('phase0', ['phase1'])]

        with self.assertRaises(Exception) as cm:
            self.try_model('pair_fixed_t_initial_below', kwargs, conns)

        msg = ("'traj' <class Trajectory>: Fixed t_initial of 5.0 is outside of allowed bounds "
               "(10.0, 15.0) for phase 'phase1'.")
        self.assertEqual(cm.exception.args[0], msg)

    def test_pair_fixed_t_initial_above(self):
        kwargs = {
            'phase0': {
                'fix_initial': True,
                'initial_val': 5.,
                'initial_bounds': (None, None),
                'fix_duration': False,
                'duration_bounds': (5., 10.),
            },
            'phase1': {
                'fix_initial': True,
                'initial_val': 99.,
                'initial_bounds': (None, None),
                'fix_duration': False,
                'duration_bounds': (5., 10.),
            }
        }
        conns = [('phase0', ['phase1'])]

        with self.assertRaises(Exception) as cm:
            self.try_model('pair_fixed_t_initial_above', kwargs, conns)

        msg = ("'traj' <class Trajectory>: Fixed t_initial of 99.0 is outside of allowed bounds "
               "(10.0, 15.0) for phase 'phase1'.")
        self.assertEqual(cm.exception.args[0], msg)

    def test_pair_t_initial_bounds_below(self):
        kwargs = {
            'phase0': {
                'initial_bounds': (3., 7.),
                'fix_duration': False,
                'duration_bounds': (5., 10.),
            },
            'phase1': {
                'initial_bounds': (5., 7.),
                'fix_duration': False,
                'duration_bounds': (5., 10.),
            }
        }
        conns = [('phase0', ['phase1'])]

        with self.assertRaises(Exception) as cm:
            self.try_model('pair_t_initial_bounds_below', kwargs, conns)

        msg = ("'traj' <class Trajectory>: t_initial bounds of (5.0, 7.0) do not overlap with "
               "allowed bounds (8.0, 17.0) for phase 'phase1'.")
        self.assertEqual(cm.exception.args[0], msg)

    def test_pair_t_initial_bounds_above(self):
        kwargs = {
            'phase0': {
                'initial_bounds': (3., 7.),
                'fix_duration': False,
                'duration_bounds': (5., 10.),
            },
            'phase1': {
                'initial_bounds': (20., 22.),
                'fix_duration': False,
                'duration_bounds': (5., 10.),
            }
        }
        conns = [('phase0', ['phase1'])]

        with self.assertRaises(Exception) as cm:
            self.try_model('pair_t_initial_bounds_above', kwargs, conns)

        msg = ("'traj' <class Trajectory>: t_initial bounds of (20.0, 22.0) do not overlap with "
               "allowed bounds (8.0, 17.0) for phase 'phase1'.")
        self.assertEqual(cm.exception.args[0], msg)

    def test_pair_no_duration_bounds(self):
        kwargs = {
            'phase0': {
                'initial_bounds': (10., 15.),
                'fix_duration': False,
            },
            'phase1': {
                'initial_bounds': (5., 7.),
                'fix_duration': False,
            }
        }
        conns = [('phase0', ['phase1'])]

        # no exception should be raised
        self.try_model('pair_t_initial_bounds', kwargs, conns)

    def test_all_fixed_t_initial(self):
        nphases = 3

        kwargs = {}
        conns = []
        for i in range(nphases):
            pname = f'phase{i}'
            kwargs[pname] = dct = {}
            dct['fix_initial'] = True
            dct['fix_duration'] = False
            dct['initial_val'] = 5. * i if i < nphases - 1 else 5. * (i - 1)
            dct['duration_bounds'] = (5. * i, 5 * (i + 1))

            if i > 0:
                conns.append((f'phase{i-1}', [pname]))

        with self.assertRaises(Exception) as cm:
            self.try_model('all_fixed', kwargs, conns)

        msg = ("'traj' <class Trajectory>: Fixed t_initial of 5.0 is outside of allowed "
               "bounds (10.0, 15.0) for phase 'phase2'.")
        self.assertEqual(cm.exception.args[0], msg)

    def test_all_t_initial_bounds(self):
        nphases = 3

        kwargs = {}
        conns = []
        for i in range(nphases):
            pname = f'phase{i}'
            kwargs[pname] = dct = {}
            dct['fix_duration'] = False
            t = 5. * i if i < nphases - 1 else 99.
            dct['initial_bounds'] = (t, t + 5.)
            dct['duration_bounds'] = (5. * i, 5 * (i + 1))

            if i > 0:
                conns.append((f'phase{i-1}', [pname]))

        with self.assertRaises(Exception) as cm:
            self.try_model('all_t_initial_bounds', kwargs, conns)

        msg = ("'traj' <class Trajectory>: t_initial bounds of (99.0, 104.0) do not overlap with "
               "allowed bounds (10.0, 20.0) for phase 'phase2'.")
        self.assertEqual(cm.exception.args[0], msg)

    def test_odd_fixed_t_initial(self):
        nphases = 4

        kwargs = {}
        conns = []
        for i in range(nphases):
            pname = f'phase{i}'
            kwargs[pname] = dct = {}
            dct['fix_initial'] = i % 2 != 0
            dct['fix_duration'] = False
            dct['initial_val'] = 5. * i if i < nphases - 1 else 5. * (i-1)
            dct['duration_bounds'] = (5. * i, 5 * (i + 1))

            if i > 0:
                conns.append((f'phase{i - 1}', [pname]))

        msg = ("'traj' <class Trajectory>: Fixed t_initial of 10.0 is outside of allowed "
               "bounds (20.0, 30.0) for phase 'phase3'.")

        with self.assertRaises(Exception) as cm:
            self.try_model('odd_fixed', kwargs, conns)

        self.assertEqual(cm.exception.args[0], msg)

    def test_odd_t_initial_bounds(self):
        nphases = 4

        kwargs = {}
        conns = []
        for i in range(nphases):
            pname = f'phase{i}'
            kwargs[pname] = dct = {}
            if i % 2 != 0:
                t = 5. * i if i < nphases - 1 else 99.
                dct['initial_bounds'] = (t, t + 5.)
            dct['fix_duration'] = False
            dct['duration_bounds'] = (5. * i, 5 * (i + 1))

            if i > 0:
                conns.append((f'phase{i - 1}', [pname]))

        msg = ("'traj' <class Trajectory>: t_initial bounds of (99.0, 104.0) do not overlap with "
               "allowed bounds (20.0, 35.0) for phase 'phase3'.")

        with self.assertRaises(Exception) as cm:
            self.try_model('odd_t_initial_bounds', kwargs, conns)

        self.assertEqual(cm.exception.args[0], msg)

    def test_even_fixed_t_initial(self):
        nphases = 5

        kwargs = {}
        conns = []
        for i in range(nphases):
            pname = f'phase{i}'
            kwargs[pname] = dct = {}
            dct['fix_initial'] = i % 2 == 0
            dct['fix_duration'] = False
            dct['initial_val'] = 5. * i
            dct['duration_bounds'] = (5. * i, 5 * (i + 1))

            if i > 0:
                conns.append((f'phase{i - 1}', [pname]))

        with self.assertRaises(Exception) as cm:
            self.try_model('even_fixed', kwargs, conns)

        msg = ("'traj' <class Trajectory>: "
               "Fixed t_initial of 20.0 is outside of allowed bounds (35.0, 45.0) for phase "
               "'phase4'.")

        self.assertEqual(cm.exception.args[0], msg)

    def test_even_t_initial_bounds(self):
        nphases = 3

        kwargs = {}
        conns = []
        for i in range(nphases):
            pname = f'phase{i}'
            kwargs[pname] = dct = {}
            if i % 2 == 0:
                t = 5. * i if i < nphases - 1 else 99.
                dct['initial_bounds'] = (t, t + 5.)
            dct['fix_duration'] = False
            dct['duration_bounds'] = (5. * i, 5 * (i + 1))

            if i > 0:
                conns.append((f'phase{i - 1}', [pname]))

        msg = ("'traj' <class Trajectory>: t_initial bounds of (99.0, 104.0) do not overlap with "
               "allowed bounds (5.0, 20.0) for phase 'phase2'.")

        with self.assertRaises(Exception) as cm:
            self.try_model('even_t_initial_bounds', kwargs, conns)

        self.assertEqual(cm.exception.args[0], msg)

    def test_branching_all_fixed_t_initial(self):
        nphases = 3  # number of phases in trunk and each branch
        nbranches = 2

        kwargs = {}
        conns = []
        branches = [f'br{i}' for i in range(nbranches)]

        # trunk phases
        for i in range(nphases):
            pname = f'phase{i}'
            kwargs[pname] = dct = {}
            dct['fix_initial'] = True
            dct['fix_duration'] = False
            dct['initial_val'] = 5.*i
            dct['duration_bounds'] = (5, 10)

            if i > 0:
                conns.append((f'phase{i-1}', [pname]))

        last_trunk_phase = pname
        last_t_min = dct['initial_val'] + dct['duration_bounds'][0]

        for br in branches:
            for i in range(nphases):
                pname = f'{br}_phase{i}'
                kwargs[pname] = dct = {}
                dct['fix_initial'] = True
                dct['fix_duration'] = False
                dct['initial_val'] = last_t_min + 5.*(i-1)
                dct['duration_bounds'] = (5, 10)

                if i > 0:
                    conns.append((f'{br}_phase{i-1}', [pname]))

            # connect branch to trunk
            conns.append((last_trunk_phase, [f'{br}_phase0']))

        with self.assertRaises(Exception) as cm:
            self.try_model('branching_all_fixed', kwargs, conns)

        msg = ("'traj' <class Trajectory>: Fixed t_initial of 10.0 is outside of allowed bounds "
               "(15.0, 20.0) for phase 'br0_phase0'.\n"
               "Fixed t_initial of 10.0 is outside of allowed bounds (15.0, 20.0) for phase "
               "'br1_phase0'.")
        self.assertEqual(cm.exception.args[0], msg)

    def test_branching_all_t_initial_bounds(self):
        nphases = 3  # number of phases in trunk and each branch
        nbranches = 2

        kwargs = {}
        conns = []
        branches = [f'br{i}' for i in range(nbranches)]

        # trunk phases
        for i in range(nphases):
            pname = f'phase{i}'
            kwargs[pname] = dct = {}
            dct['fix_duration'] = False
            dct['initial_bounds'] = (5. * i, 5. * (i + 1))
            dct['duration_bounds'] = (5, 10)

            if i > 0:
                conns.append((f'phase{i-1}', [pname]))

        last_trunk_phase = pname
        last_t_min = dct['initial_bounds'][0] + dct['duration_bounds'][0]

        for br in branches:
            for i in range(nphases):
                pname = f'{br}_phase{i}'
                kwargs[pname] = dct = {}
                dct['fix_duration'] = False
                if i == 1:
                    dct['initial_bounds'] = (0., 5)
                else:
                    t = last_t_min + 5.*(i-1)
                    dct['initial_bounds'] = (t, t + 10.)
                dct['duration_bounds'] = (5, 10)

                if i > 0:
                    conns.append((f'{br}_phase{i-1}', [pname]))

            # connect branch to trunk
            conns.append((last_trunk_phase, [f'{br}_phase0']))

        with self.assertRaises(Exception) as cm:
            self.try_model('branching_all_t_initial_bounds', kwargs, conns)

        msg = ("'traj' <class Trajectory>: t_initial bounds of (0.0, 5) do not overlap with "
               "allowed bounds (20.0, 30.0) for phase 'br0_phase1'.\n"
               "t_initial bounds of (0.0, 5) do not overlap with allowed bounds (20.0, 30.0) "
               "for phase 'br1_phase1'.")
        self.assertEqual(cm.exception.args[0], msg)

    def test_branching_odd_fixed_t_initial(self):
        nphases = 3  # number of phases in trunk and each branch
        nbranches = 2
        nbrphases = 4

        kwargs = {}
        conns = []
        branches = [f'br{i}' for i in range(nbranches)]

        # trunk phases
        for i in range(nphases):
            pname = f'phase{i}'
            kwargs[pname] = dct = {}
            dct['fix_initial'] = i % 2 != 0
            dct['fix_duration'] = False
            dct['initial_val'] = 5.*i
            dct['duration_bounds'] = (5, 10)

            if i > 0:
                conns.append((f'phase{i-1}', [pname]))

        last_trunk_phase = pname
        last_t_min = dct['initial_val'] + dct['duration_bounds'][0]

        for br in branches:
            for i in range(nbrphases):
                pname = f'{br}_phase{i}'
                kwargs[pname] = dct = {}
                dct['fix_initial'] = i % 2 != 0
                dct['fix_duration'] = False
                dct['initial_val'] = last_t_min + 15.*i
                dct['duration_bounds'] = (5, 10)

                if i > 0:
                    conns.append((f'{br}_phase{i-1}', [pname]))

            # connect branch to trunk
            conns.append((last_trunk_phase, [f'{br}_phase0']))

        with self.assertRaises(Exception) as cm:
            self.try_model('branching_odd_fixed', kwargs, conns)

        msg = ("'traj' <class Trajectory>: Fixed t_initial of 60.0 is outside of allowed bounds "
               "(40.0, 50.0) for phase 'br0_phase3'.\n"
               "Fixed t_initial of 60.0 is outside of allowed bounds (40.0, 50.0) for phase "
               "'br1_phase3'.")
        self.assertEqual(cm.exception.args[0], msg)



import numpy as np

import openmdao.api as om

import dymos as dm
from dymos.examples.finite_burn_orbit_raise.finite_burn_eom import FiniteBurnODE
from dymos.utils.misc import om_version


def make_traj(transcription='gauss-lobatto', transcription_order=3, compressed=False,
              connected=False, default_nonlinear_solver=None, default_linear_solver=None):
    """
    Build a trajectory for the finite burn orbit raise problem.

    Parameters
    ----------
    transcription : str
        One of 'radau' or 'gauss-lobatto'.
    transcription_order : int
        The order of the state transcription polynomials.
    compressed : bool
        If True, use a compressed transcription.
    connected : bool
        If True, connect the phases together, otherwise enforce continuity via constraints.

    Returns
    -------
    traj : dm.Trajectory
        The Trajectory object.
    """
    t = {'gauss-lobatto': dm.GaussLobatto(num_segments=5, order=transcription_order, compressed=compressed),
         'radau': dm.Radau(num_segments=5, order=transcription_order, compressed=compressed)}

    traj = dm.Trajectory()

    if default_nonlinear_solver is not None:
        traj.phases.nonlinear_solver = default_nonlinear_solver

    if default_linear_solver is not None:
        traj.phases.linear_solver = default_linear_solver

    traj.add_parameter('c', opt=False, val=1.5, units='DU/TU',
                       targets={'burn1': ['c'], 'burn2': ['c']})

    # First Phase (burn)

    burn1 = dm.Phase(ode_class=FiniteBurnODE, transcription=t[transcription])

    burn1 = traj.add_phase('burn1', burn1)

    burn1.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='TU')
    burn1.add_state('r', fix_initial=True, fix_final=False, defect_scaler=1.0,
                    rate_source='r_dot', units='DU')
    burn1.add_state('theta', fix_initial=True, fix_final=False, defect_scaler=1.0E-3,
                    rate_source='theta_dot', units='rad')
    burn1.add_state('vr', fix_initial=True, fix_final=False, defect_scaler=1.0,
                    rate_source='vr_dot', units='DU/TU')
    burn1.add_state('vt', fix_initial=True, fix_final=False, defect_scaler=1.0,
                    rate_source='vt_dot', units='DU/TU')
    burn1.add_state('accel', fix_initial=True, fix_final=False,
                    rate_source='at_dot', units='DU/TU**2')
    burn1.add_state('deltav', fix_initial=True, fix_final=False,
                    rate_source='deltav_dot', units='DU/TU')
    burn1.add_control('u1', rate_continuity=False, rate2_continuity=False, units='deg', scaler=0.01,
                      rate_continuity_scaler=0.001, rate2_continuity_scaler=0.001,
                      lower=-30, upper=30)
    # Second Phase (Coast)
    coast = dm.Phase(ode_class=FiniteBurnODE, transcription=t[transcription])

    coast.set_time_options(initial_bounds=(0.5, 20), duration_bounds=(.5, 50), duration_ref=50, units='TU')
    coast.add_state('r', fix_initial=False, fix_final=False, defect_scaler=1.0,
                    rate_source='r_dot', units='DU')
    coast.add_state('theta', fix_initial=False, fix_final=False, defect_scaler=1.0E-3,
                    rate_source='theta_dot', units='rad')
    coast.add_state('vr', fix_initial=False, fix_final=False, defect_scaler=1.0,
                    rate_source='vr_dot', units='DU/TU')
    coast.add_state('vt', fix_initial=False, fix_final=False, defect_scaler=1.0,
                    rate_source='vt_dot', units='DU/TU')
    coast.add_state('accel', fix_initial=True, fix_final=True,
                    rate_source='at_dot', units='DU/TU**2')
    coast.add_state('deltav', fix_initial=False, fix_final=False,
                    rate_source='deltav_dot', units='DU/TU')

    coast.add_parameter('u1', opt=False, val=0.0, units='deg')
    coast.add_parameter('c', val=1.0, units='DU/TU')

    # Third Phase (burn)
    burn2 = dm.Phase(ode_class=FiniteBurnODE, transcription=t[transcription])

    traj.add_phase('coast', coast)
    traj.add_phase('burn2', burn2)

    if connected:

        burn2.set_time_options(initial_bounds=(1.0, 60), duration_bounds=(-10.0, -0.5),
                               initial_ref=10, units='TU')
        burn2.add_state('r', fix_initial=True, fix_final=False, defect_scaler=1.0E-2,
                        rate_source='r_dot', units='DU')
        burn2.add_state('theta', fix_initial=False, fix_final=False, defect_scaler=1000.0,
                        rate_source='theta_dot', units='rad')
        burn2.add_state('vr', fix_initial=True, fix_final=False, defect_scaler=1.0E-2,
                        rate_source='vr_dot', units='DU/TU')
        burn2.add_state('vt', fix_initial=True, fix_final=False, defect_scaler=1.0,
                        rate_source='vt_dot', units='DU/TU')
        burn2.add_state('accel', fix_initial=False, fix_final=False, defect_scaler=1.0,
                        rate_source='at_dot', units='DU/TU**2')
        burn2.add_state('deltav', fix_initial=False, fix_final=False, defect_scaler=1.0,
                        rate_source='deltav_dot', units='DU/TU')

        burn2.add_objective('deltav', loc='initial', scaler=100.0)

        burn2.add_control('u1', rate_continuity=False, rate2_continuity=False, units='deg',
                          scaler=0.01, lower=-180, upper=180)
    else:

        burn2.set_time_options(initial_bounds=(0.5, 50), duration_bounds=(.5, 10), initial_ref=10, units='TU')
        burn2.add_state('r', fix_initial=False, fix_final=True, defect_scaler=1.0E-2,
                        rate_source='r_dot', targets=['r'], units='DU')
        burn2.add_state('theta', fix_initial=False, fix_final=False, defect_scaler=1000.0,
                        rate_source='theta_dot', targets=['theta'], units='rad')
        burn2.add_state('vr', fix_initial=False, fix_final=True, defect_scaler=1.0E-2,
                        rate_source='vr_dot', targets=['vr'], units='DU/TU')
        burn2.add_state('vt', fix_initial=False, fix_final=True, defect_scaler=1.0,
                        rate_source='vt_dot', targets=['vt'], units='DU/TU')
        burn2.add_state('accel', fix_initial=False, fix_final=False, defect_scaler=1.0,
                        rate_source='at_dot', targets=['accel'], units='DU/TU**2')
        burn2.add_state('deltav', fix_initial=False, fix_final=False, lower=0.0, defect_scaler=1.0,
                        rate_source='deltav_dot', units='DU/TU')

        burn2.add_objective('deltav', loc='final', scaler=100.0)

        burn2.add_control('u1', rate_continuity=False, rate2_continuity=False, units='deg',
                          scaler=0.01, lower=-180, upper=180, targets=['u1'])

    burn1.add_timeseries_output('pos_x')
    coast.add_timeseries_output('pos_x')
    burn2.add_timeseries_output('pos_x')

    burn1.add_timeseries_output('pos_y')
    coast.add_timeseries_output('pos_y')
    burn2.add_timeseries_output('pos_y')

    # Link Phases
    if connected:
        traj.link_phases(phases=['burn1', 'coast'],
                         vars=['time', 'r', 'theta', 'vr', 'vt', 'deltav'],
                         connected=True)

        # No direct connections to the end of a phase.
        traj.link_phases(phases=['burn2', 'coast'],
                         vars=['r', 'theta', 'vr', 'vt', 'deltav'],
                         locs=('final', 'final'))
        traj.link_phases(phases=['burn2', 'coast'],
                         vars=['time'], locs=('final', 'final'))

        traj.link_phases(phases=['burn1', 'burn2'], vars=['accel'],
                         locs=('final', 'final'))

    else:
        traj.link_phases(phases=['burn1', 'coast', 'burn2'],
                         vars=['time', 'r', 'theta', 'vr', 'vt', 'deltav'])

        traj.link_phases(phases=['burn1', 'burn2'], vars=['accel'])

    return traj


def two_burn_orbit_raise_problem(transcription='gauss-lobatto', optimizer='SLSQP', r_target=3.0,
                                 transcription_order=3, compressed=False, run_driver=True,
                                 max_iter=300, simulate=True, show_output=True, connected=False, restart=None,
                                 solution_record_file='dymos_solution.db', simulation_record_file='dymos_simulation.db',
                                 default_nonlinear_solver=None, default_linear_solver=None):
    """
    Build and run the finite burn orbit raise problem.

    Parameters
    ----------
    transcription : str
        One of 'radau' or 'gauss-lobatto'.
    optimizer : str
        The pyoptsparse optimizer to use.
    r_target : float
        The final radius target, in AU.
    transcription_order : int
        The order of the state transcription.
    compressed : bool
        If True, use a compressed transcription.
    run_driver : bool
        If True, run the optimization driver.
    max_iter : int
        The maximum allowable number of optimizer iterations.
    simulate : bool
        If True, run simulate on the solution.
    show_output : bool
        If True, display the optimizer output.
    connected : bool
        If True, connect phases for continuity.  Otherwise, enforce continuity via linkage constraints.
    restart : str or None
        The restart file to use, if available.

    Returns
    -------
    p : om.Problem
        The OpenMDAO Problem instance.
    """
    p = om.Problem(model=om.Group())

    p.driver = om.pyOptSparseDriver()
    p.driver.options['optimizer'] = optimizer
    p.driver.declare_coloring()
    if optimizer == 'SNOPT':
        p.driver.opt_settings['Major iterations limit'] = max_iter
        p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
        p.driver.opt_settings['Major optimality tolerance'] = 1.0E-6
        if show_output:
            p.driver.opt_settings['iSumm'] = 6
    elif optimizer == 'IPOPT':
        p.driver.opt_settings['mu_init'] = 1e-3
        p.driver.opt_settings['max_iter'] = max_iter
        p.driver.opt_settings['acceptable_tol'] = 1e-3
        p.driver.opt_settings['constr_viol_tol'] = 1e-3
        p.driver.opt_settings['compl_inf_tol'] = 1e-3
        p.driver.opt_settings['acceptable_iter'] = 0
        p.driver.opt_settings['tol'] = 1e-3
        p.driver.opt_settings['print_level'] = 5 if show_output else 0
        p.driver.opt_settings['nlp_scaling_method'] = 'gradient-based'  # for faster convergence
        p.driver.opt_settings['alpha_for_y'] = 'safer-min-dual-infeas'
        p.driver.opt_settings['mu_strategy'] = 'monotone'
        p.driver.opt_settings['derivative_test'] = 'first-order'

    traj = make_traj(transcription=transcription, transcription_order=transcription_order,
                     compressed=compressed, connected=connected,
                     default_nonlinear_solver=default_nonlinear_solver,
                     default_linear_solver=default_linear_solver)
    p.model.add_subsystem('traj', subsys=traj)

    # Needed to move the direct solver down into the phases for use with MPI.
    #  - After moving down, used fewer iterations (about 30 less)

    p.setup(check=False)

    # Set Initial Guesses
    traj.set_parameter_val('c', val=1.5, units='DU/TU')

    burn1 = p.model._get_subsystem('traj.phases.burn1')
    burn2 = p.model._get_subsystem('traj.phases.burn2')
    coast = p.model._get_subsystem('traj.phases.coast')

    if burn1 in p.model.traj.phases._subsystems_myproc:
        burn1.set_time_val(initial=0.0, duration=2.25)
        burn1.set_state_val('r', [1, 1.5])
        burn1.set_state_val('theta', [0, 1.7])
        burn1.set_state_val('vr', [0, 0])
        burn1.set_state_val('vt', [1, 1])
        burn1.set_state_val('accel', [0.1, 0.0])
        burn1.set_state_val('deltav', [0, 0.1])
        burn1.set_control_val('u1', [0.0, 0.0])

    if coast in p.model.traj.phases._subsystems_myproc:
        coast.set_time_val(initial=2.25, duration=3.0)
        coast.set_state_val('r', [1.3, 1.5])
        coast.set_state_val('theta', [2.1767, 1.7])
        coast.set_state_val('vr', [0.3285, 0])
        coast.set_state_val('vt', [0.97, 1])
        coast.set_state_val('accel', [0, 0])
        coast.set_parameter_val('u1', 0)

    if burn2 in p.model.traj.phases._subsystems_myproc:
        if connected:
            burn2.set_time_val(initial=7.0, duration=-1.75)
            burn2.set_state_val('r', [r_target, 1])
            burn2.set_state_val('theta', [4, 0])
            burn2.set_state_val('vr', [0, 0])
            burn2.set_state_val('vt', [np.sqrt(1 / r_target), 1])
            burn2.set_state_val('accel', [0.0, 0.1])
            burn2.set_state_val('deltav', [0.2, 0.1])
            burn2.set_control_val('u1', [0, 0])

        else:
            burn2.set_time_val(initial=5.25, duration=1.75)
            burn2.set_state_val('r', [1, r_target])
            burn2.set_state_val('theta', [0, 4])
            burn2.set_state_val('vr', [0, 0])
            burn2.set_state_val('vt', [1, np.sqrt(1 / r_target)])
            burn2.set_state_val('accel', [0.1, 0.0])
            burn2.set_state_val('deltav', [0.1, 0.2])
            burn2.set_control_val('u1', [0, 0])

    p.final_setup()
    if burn2 in p.model.traj.phases._subsystems_myproc:
        print(burn2.get_val('t_initial'))
        print(burn2.get_val('t_duration'))

    if run_driver or simulate:
        dm.run_problem(p, run_driver=run_driver, simulate=simulate, restart=restart, make_plots=True,
                       solution_record_file=solution_record_file, simulation_record_file=simulation_record_file)

    return p


if __name__ == '__main__':  # pragma: no cover

    import openmdao.api as om
    from openmdao.utils.assert_utils import assert_near_equal

    optimizer = 'IPOPT'

    CONNECTED = False

    p = two_burn_orbit_raise_problem(transcription='gauss-lobatto', transcription_order=3,
                                     compressed=False, optimizer=optimizer, simulate=True,
                                     connected=CONNECTED, show_output=False)

    sol_db = 'dymos_solution.db'
    sim_db = 'dymos_simulation.db'
    if om_version()[0] > (3, 34, 2):
        sol_db = p.get_outputs_dir() / sol_db
        sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

    sol_case = om.CaseReader(sol_db).get_case('final')
    sim_case = om.CaseReader(sim_db).get_case('final')

    # The last phase in this case is run in reverse time if CONNECTED=True,
    # so grab the correct index to test the resulting delta-V.
    end_idx = 0 if CONNECTED else -1

    assert_near_equal(sol_case.get_val('traj.burn2.timeseries.deltav')[end_idx], 0.3995, tolerance=2.0E-3)
    assert_near_equal(sim_case.get_val('traj.burn2.timeseries.deltav')[end_idx], 0.3995, tolerance=2.0E-3)



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE


class BrachistochroneArclengthODE(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input('v', val=np.zeros(nn), desc='velocity', units='m/s')
        self.add_input('theta', val=np.zeros(nn), desc='angle of wire', units='rad')
        self.add_output('Sdot', val=np.zeros(nn), desc='rate of change of arclength', units='m/s')

        # Setup partials
        arange = np.arange(self.options['num_nodes'])

        self.declare_partials(of='Sdot', wrt='v', rows=arange, cols=arange)
        self.declare_partials(of='Sdot', wrt='theta', rows=arange, cols=arange)

    def compute(self, inputs, outputs):
        theta = inputs['theta']
        v = inputs['v']
        outputs['Sdot'] = np.sqrt(1.0 + (1.0/np.tan(theta))**2) * v * np.sin(theta)

    def compute_partials(self, inputs, jacobian):
        theta = inputs['theta']
        v = inputs['v']
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)
        tan_theta = np.tan(theta)
        cot_theta = 1.0 / tan_theta
        csc_theta = 1.0 / sin_theta

        jacobian['Sdot', 'v'] = sin_theta * np.sqrt(1.0 + cot_theta**2)
        jacobian['Sdot', 'theta'] = v * (cos_theta * (cot_theta**2 + 1) - cot_theta * csc_theta) / \
            (np.sqrt(1 + cot_theta**2))


def make_brachistochrone_phase(transcription='gauss-lobatto', num_segments=8, transcription_order=3,
                               compressed=True):

    if transcription == 'gauss-lobatto':
        t = dm.GaussLobatto(num_segments=num_segments,
                            order=transcription_order,
                            compressed=compressed)
    elif transcription == 'radau-ps':
        t = dm.Radau(num_segments=num_segments,
                     order=transcription_order,
                     compressed=compressed)

    phase = dm.Phase(ode_class=BrachistochroneODE, transcription=t)

    return phase


@use_tempdirs
@require_pyoptsparse(optimizer='SLSQP')
class TestTandemPhases(unittest.TestCase):

    def _run_transcription(self, transcription):
        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring()

        #
        # First Phase: Standard Brachistochrone
        #
        num_segments = 10
        transcription_order = 3
        compressed = False

        if transcription == 'gauss-lobatto':
            tx0 = dm.GaussLobatto(num_segments=num_segments,
                                  order=transcription_order,
                                  compressed=compressed)
            tx1 = dm.GaussLobatto(num_segments=num_segments*2, order=transcription_order*3,
                                  compressed=compressed)

        elif transcription == 'radau-ps':
            tx0 = dm.Radau(num_segments=num_segments,
                           order=transcription_order,
                           compressed=compressed)
            tx1 = dm.Radau(num_segments=num_segments*2,
                           order=transcription_order*3,
                           compressed=compressed)

        phase0 = dm.Phase(ode_class=BrachistochroneODE, transcription=tx0)

        p.model.add_subsystem('phase0', phase0)

        phase0.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase0.add_state('x', fix_initial=True, fix_final=False)
        phase0.add_state('y', fix_initial=True, fix_final=False)
        phase0.add_state('v', fix_initial=True)

        phase0.add_parameter('g', units='m/s**2', val=9.80665)

        phase0.add_control('theta', units='deg',
                           rate_continuity=False, lower=0.01, upper=179.9)

        phase0.add_boundary_constraint('x', loc='final', equals=10)
        phase0.add_boundary_constraint('y', loc='final', equals=5)

        # Add alternative timeseries output to provide control inputs for the next phase
        phase0.add_timeseries('timeseries2', transcription=tx1, subset='control_input')

        #
        # Second Phase: Integration of ArcLength
        #

        phase1 = dm.Phase(ode_class=BrachistochroneArclengthODE, transcription=tx1)

        p.model.add_subsystem('phase1', phase1)

        phase1.set_time_options(fix_initial=True, input_duration=True)

        phase1.add_state('S', fix_initial=True, fix_final=False, rate_source='Sdot')

        phase1.add_control('theta', opt=False, units='deg', targets=['theta'])
        phase1.add_control('v', opt=False, units='m/s', targets=['v'])

        #
        # Connect the two phases
        #
        p.model.connect('phase0.t_duration_val', 'phase1.t_duration')

        p.model.connect('phase0.timeseries2.theta', 'phase1.controls:theta')
        p.model.connect('phase0.timeseries2.v', 'phase1.controls:v')

        # Minimize time at the end of the phase
        # phase1.add_objective('time', loc='final', scaler=1)
        # phase1.add_boundary_constraint('S', loc='final', upper=12)

        phase1.add_objective('S', loc='final', ref=1)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True)

        phase0.set_time_val(initial=0.0, duration=2.0)

        phase0.set_state_val('x', [0, 10])
        phase0.set_state_val('y', [10, 5])
        phase0.set_state_val('v', [0, 9.9])

        phase0.set_control_val('theta', [5, 100])
        phase0.set_parameter_val('g', 9.80665)

        phase1.set_state_val('S', 0.0)

        p.run_driver()

        expected = np.sqrt((10-0)**2 + (10 - 5)**2)
        assert_near_equal(p['phase1.timeseries.S'][-1], expected, tolerance=1.0E-3)

    def test_tandem_phases_radau(self):
        self._run_transcription('radau-ps')

    def test_tandem_phases_gl(self):
        self._run_transcription('gauss-lobatto')


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

import numpy as np
import openmdao.api as om
import dymos as dm
from dymos import options as dymos_options
from dymos.utils.misc import om_version

from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal
from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE
from dymos.transcriptions.explicit_shooting.ode_integration_comp import ODEIntegrationComp

from dymos.phase.options import TimeOptionsDictionary, StateOptionsDictionary, ParameterOptionsDictionary
from dymos.transcriptions.grid_data import GridData
from dymos.utils.testing_utils import SimpleODE


class TestODEIntegrationComp(unittest.TestCase):

    @unittest.skipIf(om_version()[0] >= (3, 36, 0) or om_version()[0] < (3, 37, 0),
                     reason='Test skipped due to an issue in OpenMDAO 3.36.x')
    def test_integrate_scalar_ode(self):
        dymos_options['include_check_partials'] = True

        for ogd_eq_igd in (True, False):
            with self.subTest('input_grid_data = output_grid_data'
                              if ogd_eq_igd else 'input_grid_data != output_grid_data'):

                input_grid_data = GridData(num_segments=1, transcription='gauss-lobatto', transcription_order=3)

                if ogd_eq_igd:
                    output_grid_data = input_grid_data
                else:
                    output_grid_data = GridData(num_segments=1, transcription='uniform', transcription_order=10)

                time_options = TimeOptionsDictionary()

                time_options['targets'] = 't'
                time_options['units'] = 's'

                state_options = {'x': StateOptionsDictionary()}

                state_options['x']['shape'] = (1,)
                state_options['x']['units'] = 's**2'
                state_options['x']['rate_source'] = 'x_dot'
                state_options['x']['targets'] = ['x']

                param_options = {'p': ParameterOptionsDictionary()}

                param_options['p']['shape'] = (1,)
                param_options['p']['units'] = 's**2'
                param_options['p']['targets'] = ['p']

                control_options = {}

                prob = om.Problem()

                prob.model.add_subsystem('integrator',
                                         ODEIntegrationComp(input_grid_data=input_grid_data,
                                                            output_grid_data=output_grid_data,
                                                            time_options=time_options, state_options=state_options,
                                                            parameter_options=param_options,
                                                            control_options=control_options,
                                                            ode_class=SimpleODE, ode_init_kwargs=None))
                prob.setup()
                prob.set_val('integrator.states:x', 0.5)
                prob.set_val('integrator.t_initial', 0.0)
                prob.set_val('integrator.t_duration', 2.0)
                prob.set_val('integrator.parameters:p', 1.0)

                prob.run_model()

                x = prob.get_val('integrator.states_out:x')
                t = prob.get_val('integrator.time')
                p = prob.get_val('integrator.parameters:p')

                expected = t**2 + 2 * t + p - 0.5 * np.exp(t)

                assert_near_equal(x, expected, tolerance=1.0E-5)

                cpd = prob.check_partials(compact_print=True)
                assert_check_partials(cpd, atol=1.0E-4, rtol=1.0E-5)

        dymos_options['include_check_partials'] = False

    @unittest.skipIf(om_version()[0] >= (3, 36, 0) or om_version()[0] < (3, 37, 0),
                     reason='Test skipped due to an issue in OpenMDAO 3.36.x')
    def test_integrate_with_controls(self):

        dymos_options['include_check_partials'] = True

        gd = dm.transcriptions.grid_data.GridData(num_segments=5, transcription='gauss-lobatto',
                                                  transcription_order=3, compressed=True)

        time_options = dm.phase.options.TimeOptionsDictionary()

        time_options['units'] = 's'

        state_options = {'x': dm.phase.options.StateOptionsDictionary(),
                         'y': dm.phase.options.StateOptionsDictionary(),
                         'v': dm.phase.options.StateOptionsDictionary()}

        state_options['x']['shape'] = (1,)
        state_options['x']['units'] = 'm'
        state_options['x']['rate_source'] = 'xdot'
        state_options['x']['targets'] = []

        state_options['y']['shape'] = (1,)
        state_options['y']['units'] = 'm'
        state_options['y']['rate_source'] = 'ydot'
        state_options['y']['targets'] = []

        state_options['v']['shape'] = (1,)
        state_options['v']['units'] = 'm/s'
        state_options['v']['rate_source'] = 'vdot'
        state_options['v']['targets'] = ['v']

        param_options = {'g': dm.phase.options.ParameterOptionsDictionary()}

        param_options['g']['shape'] = (1,)
        param_options['g']['units'] = 'm/s**2'
        param_options['g']['targets'] = ['g']

        control_options = {'theta': dm.phase.options.ControlOptionsDictionary()}

        control_options['theta']['shape'] = (1,)
        control_options['theta']['units'] = 'rad'
        control_options['theta']['targets'] = ['theta']

        p = om.Problem()

        p.model.add_subsystem('integrator',
                              ODEIntegrationComp(ode_class=BrachistochroneODE,
                                                 time_options=time_options,
                                                 state_options=state_options,
                                                 parameter_options=param_options,
                                                 control_options=control_options,
                                                 input_grid_data=gd,
                                                 ode_init_kwargs=None))

        p.setup()

        p.set_val('integrator.states:x', 0.0)
        p.set_val('integrator.states:y', 10.0)
        p.set_val('integrator.states:v', 0.0)
        p.set_val('integrator.t_initial', 0.0)
        p.set_val('integrator.t_duration', 1.8016)
        p.set_val('integrator.parameters:g', 9.80665)

        p.set_val('integrator.controls:theta', np.linspace(0.01, 100.0, gd.subset_num_nodes['control_input']),
                  units='deg')

        p.run_model()

        x = p.get_val('integrator.states_out:x')
        y = p.get_val('integrator.states_out:y')
        v = p.get_val('integrator.states_out:v')

        # These tolerances are loose since theta is not properly spaced along the lgl nodes.
        assert_near_equal(x[-1, ...], 10.0, tolerance=0.1)
        assert_near_equal(y[-1, ...], 5.0, tolerance=0.1)
        assert_near_equal(v[-1, ...], 9.9, tolerance=0.1)

        with np.printoptions(linewidth=1024):
            cpd = p.check_partials(compact_print=False, method='fd')
            assert_check_partials(cpd, atol=1.0E-4, rtol=1.0E-4)

    @unittest.skipIf(om_version()[0] >= (3, 36, 0) or om_version()[0] < (3, 37, 0),
                     reason='Test skipped due to an issue in OpenMDAO 3.36.x')
    def test_integrate_with_polynomial_controls(self):

        dymos_options['include_check_partials'] = True

        gd = dm.transcriptions.grid_data.GridData(num_segments=5, transcription='gauss-lobatto',
                                                  transcription_order=3, compressed=True)

        time_options = dm.phase.options.TimeOptionsDictionary()

        time_options['units'] = 's'

        state_options = {'x': dm.phase.options.StateOptionsDictionary(),
                         'y': dm.phase.options.StateOptionsDictionary(),
                         'v': dm.phase.options.StateOptionsDictionary()}

        state_options['x']['shape'] = (1,)
        state_options['x']['units'] = 'm'
        state_options['x']['rate_source'] = 'xdot'
        state_options['x']['targets'] = []

        state_options['y']['shape'] = (1,)
        state_options['y']['units'] = 'm'
        state_options['y']['rate_source'] = 'ydot'
        state_options['y']['targets'] = []

        state_options['v']['shape'] = (1,)
        state_options['v']['units'] = 'm/s'
        state_options['v']['rate_source'] = 'vdot'
        state_options['v']['targets'] = ['v']

        param_options = {'g': dm.phase.options.ParameterOptionsDictionary()}

        param_options['g']['shape'] = (1,)
        param_options['g']['units'] = 'm/s**2'
        param_options['g']['targets'] = ['g']

        control_options = {'theta': dm.phase.options.ControlOptionsDictionary()}

        control_options['theta']['control_type'] = 'polynomial'
        control_options['theta']['shape'] = (1,)
        control_options['theta']['order'] = 2
        control_options['theta']['units'] = 'rad'
        control_options['theta']['targets'] = ['theta']

        p = om.Problem()

        p.model.add_subsystem('integrator',
                              ODEIntegrationComp(ode_class=BrachistochroneODE,
                                                 time_options=time_options,
                                                 state_options=state_options,
                                                 parameter_options=param_options,
                                                 control_options=control_options,
                                                 input_grid_data=gd,
                                                 ode_init_kwargs=None))

        p.setup()

        p.set_val('integrator.states:x', 0.0)
        p.set_val('integrator.states:y', 10.0)
        p.set_val('integrator.states:v', 0.0)
        p.set_val('integrator.t_initial', 0.0)
        p.set_val('integrator.t_duration', 1.8016)
        p.set_val('integrator.parameters:g', 9.80665)

        p.set_val('integrator.controls:theta',
                  np.linspace(0.01, 100.0, control_options['theta']['order']+1),
                  units='deg')

        p.run_model()

        x = p.get_val('integrator.states_out:x')
        y = p.get_val('integrator.states_out:y')
        v = p.get_val('integrator.states_out:v')

        # These tolerances are loose since theta is not properly spaced along the lgl nodes.
        assert_near_equal(x[-1, ...], 10.0, tolerance=0.1)
        assert_near_equal(y[-1, ...], 5.0, tolerance=0.1)
        assert_near_equal(v[-1, ...], 9.9, tolerance=0.1)

        with np.printoptions(linewidth=1024):
            cpd = p.check_partials(compact_print=False, method='fd')
            assert_check_partials(cpd, atol=1.0E-4, rtol=1.0E-4)

        dymos_options['include_check_partials'] = False


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest


import numpy as np
import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs
import dymos as dm


class _BrysonDenhamODE(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input('x', shape=(nn,), units='m')
        self.add_input('v', shape=(nn,), units='m/s')
        self.add_input('u', shape=(nn,), units='m/s**2')

        self.add_output('J_dot', shape=(nn,), units='m**2/s**4',
                        tags=['dymos.state_rate_source:J',
                              'dymos.state_units:m**2/s**3'])

        ar = np.arange(nn, dtype=int)

        self.declare_partials(of='J_dot', wrt='u', rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        u = inputs['u']
        outputs['J_dot'] = 0.5 * u ** 2

    def compute_partials(self, inputs, partials):
        partials['J_dot', 'u'] = inputs['u']


@use_tempdirs
class TestGridRefinement(unittest.TestCase):

    def test_refine_hp_non_ode_rate_sources(self):
        p = om.Problem()

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        traj = p.model.add_subsystem('traj', dm.Trajectory())
        tx = dm.Radau(num_segments=5)
        phase = traj.add_phase('phase0', dm.Phase(ode_class=_BrysonDenhamODE, transcription=tx))

        #
        # Set the variables
        #
        phase.set_time_options(fix_initial=True, fix_duration=True)

        phase.add_state('x', fix_initial=True, fix_final=True, rate_source='v')
        phase.add_state('v', fix_initial=True, fix_final=True, rate_source='u')
        phase.add_state('J', fix_initial=True, fix_final=False)
        phase.add_control('u', continuity=True, rate_continuity=False)

        #
        # Minimize time at the end of the phase
        #
        phase.add_objective('J', loc='final', ref=1)
        phase.add_path_constraint('x', upper=1/9)

        #
        # Setup the Problem
        #
        p.setup()

        #
        # Set the initial values
        #
        phase.set_time_val(initial=0, duration=1)

        phase.set_state_val('x', [0, 0])
        phase.set_state_val('v', [1, -1])
        phase.set_state_val('J', [0, 1])
        phase.set_control_val('u', [0, 0])

        #
        # Solve for the optimal trajectory
        #
        dm.run_problem(p, run_driver=True, simulate=True, refine_iteration_limit=5)

        num_seg = p.model.traj.phases.phase0.options['transcription'].grid_data.num_segments
        seg_orders = p.model.traj.phases.phase0.options['transcription'].grid_data.transcription_order

        self.assertGreaterEqual(num_seg, 5)
        self.assertGreater(sum(seg_orders), 5 * 3)

    def test_refine_ph_non_ode_rate_sources(self):
        p = om.Problem()

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        traj = p.model.add_subsystem('traj', dm.Trajectory())
        tx = dm.Radau(num_segments=5, order=3)
        phase = traj.add_phase('phase0', dm.Phase(ode_class=_BrysonDenhamODE, transcription=tx))

        #
        # Set the variables
        #
        phase.set_time_options(fix_initial=True, fix_duration=True)

        phase.add_state('x', fix_initial=True, fix_final=True, rate_source='v')
        phase.add_state('v', fix_initial=True, fix_final=True, rate_source='u')
        phase.add_state('J', fix_initial=True, fix_final=False)
        phase.add_control('u', continuity=True, rate_continuity=False)

        #
        # Minimize time at the end of the phase
        #
        phase.add_objective('J', loc='final', ref=1)
        phase.add_path_constraint('x', upper=1/9)

        #
        # Setup the Problem
        #
        p.setup()

        #
        # Set the initial values
        #
        p['traj.phase0.t_initial'] = 0.0
        p['traj.phase0.t_duration'] = 1.0

        p.set_val('traj.phase0.states:x', phase.interp('x', ys=[0, 0]))
        p.set_val('traj.phase0.states:v', phase.interp('v', ys=[1, -1]))
        p.set_val('traj.phase0.states:J', phase.interp('J', ys=[0, 1]))
        p.set_val('traj.phase0.controls:u', np.sin(phase.interp('u', ys=[0, 0])))

        #
        # Solve for the optimal trajectory
        #
        dm.run_problem(p, run_driver=True, simulate=True,
                       refine_method='ph', refine_iteration_limit=5)

        num_seg = p.model.traj.phases.phase0.options['transcription'].grid_data.num_segments
        seg_orders = p.model.traj.phases.phase0.options['transcription'].grid_data.transcription_order

        self.assertGreaterEqual(num_seg, 5)
        self.assertGreater(sum(seg_orders), 5 * 3)



from packaging.version import Version
import unittest

import numpy as np

import openmdao
import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse
from openmdao.utils.assert_utils import assert_near_equal

import dymos as dm
from dymos.examples.balanced_field.balanced_field_ode import BalancedFieldODEComp
from dymos.examples.balanced_field.balanced_field_length import make_balanced_field_length_problem
from dymos.utils.misc import om_version


@use_tempdirs
class TestBalancedFieldLengthRestart(unittest.TestCase):

    @require_pyoptsparse(optimizer='IPOPT')
    @unittest.skipUnless(Version(openmdao.__version__) > Version("3.23"),
                         reason='Test requires OpenMDAO 3.23.0 or later.')
    def test_make_plots(self):
        p = make_balanced_field_length_problem(ode_class=BalancedFieldODEComp, tx=dm.Radau(num_segments=3))
        dm.run_problem(p, run_driver=True, simulate=True, make_plots=True)

    @require_pyoptsparse(optimizer='IPOPT')
    @unittest.skipUnless(Version(openmdao.__version__) > Version("3.23"),
                         reason='Test requires OpenMDAO 3.23.0 or later.')
    def test_restart_from_sol(self):
        p = make_balanced_field_length_problem(ode_class=BalancedFieldODEComp, tx=dm.Radau(num_segments=3))
        dm.run_problem(p, run_driver=True, simulate=False)

        sol_db = 'dymos_solution.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db

        sol_results = om.CaseReader(sol_db).get_case('final')

        assert_near_equal(sol_results.get_val('traj.climb.timeseries.r')[-1], 2197, tolerance=0.01)

        dm.run_problem(p, run_driver=True, simulate=True, restart=sol_db)

        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        sol_results = om.CaseReader(sol_db).get_case('final')
        sim_results = om.CaseReader(sim_db).get_case('final')

        assert_near_equal(sol_results.get_val('traj.climb.timeseries.r')[-1], 2197, tolerance=0.01)
        assert_near_equal(sim_results.get_val('traj.climb.timeseries.r')[-1], 2197, tolerance=0.01)

        assert_near_equal(sol_results.get_val('traj.rto.timeseries.r')[-1], 2197, tolerance=0.01)
        assert_near_equal(sim_results.get_val('traj.rto.timeseries.r')[-1], 2197, tolerance=0.01)

    @require_pyoptsparse(optimizer='IPOPT')
    @unittest.skipUnless(Version(openmdao.__version__) > Version("3.23"),
                         reason='Test requires OpenMDAO 3.23.0 or later.')
    def test_restart_from_sim(self):
        p = make_balanced_field_length_problem(ode_class=BalancedFieldODEComp, tx=dm.Radau(num_segments=3))
        dm.run_problem(p, run_driver=True, simulate=True)

        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = p.get_outputs_dir() / sol_db
            sim_db = p.model.traj.sim_prob.get_outputs_dir() / sim_db

        sol_results = om.CaseReader(sol_db).get_case('final')

        assert_near_equal(sol_results.get_val('traj.climb.timeseries.r')[-1], 2197, tolerance=0.01)

        dm.run_problem(p, run_driver=True, simulate=True, restart=sim_db)

        sol_results = om.CaseReader(sol_db).get_case('final')
        sim_results = om.CaseReader(sim_db).get_case('final')

        assert_near_equal(sol_results.get_val('traj.climb.timeseries.r')[-1], 2197, tolerance=0.01)
        assert_near_equal(sim_results.get_val('traj.climb.timeseries.r')[-1], 2197, tolerance=0.01)

        assert_near_equal(sol_results.get_val('traj.rto.timeseries.r')[-1], 2197, tolerance=0.01)
        assert_near_equal(sim_results.get_val('traj.rto.timeseries.r')[-1], 2197, tolerance=0.01)


@use_tempdirs
class TestBalancedFieldLengthDefaultValues(unittest.TestCase):

    @require_pyoptsparse(optimizer='IPOPT')
    def test_default_vals_stick(self):
        """
        Make the balanced field problem without any set_val calls after setup.
        """
        p = make_balanced_field_length_problem(ode_class=BalancedFieldODEComp, tx=dm.GaussLobatto(num_segments=3))

        p.run_model()

        assert_near_equal(p.get_val('traj.rotate.t_initial'), 35)
        assert_near_equal(p.get_val('traj.rotate.t_duration'), 5)
        assert_near_equal(p.get_val('traj.rotate.controls:alpha'), np.array([[0, 0]]).T)
        assert_near_equal(p.get_val('traj.climb.controls:alpha', units='deg'),
                          p.model.traj.phases.climb.interp('', [5, 5], nodes='control_input'))
        assert_near_equal(p.get_val('traj.climb.states:gam', units='deg'),
                          p.model.traj.phases.climb.interp(ys=[0.0, 5.0], nodes='state_input'))
        assert_near_equal(p.get_val('traj.climb.states:h', units='ft'),
                          p.model.traj.phases.climb.interp(ys=[0.0, 35.0], nodes='state_input'))
        assert_near_equal(p.get_val('traj.v1_to_vr.parameters:alpha'), 0.0)



__version__ = '1.13.0'


from .phase import Phase, AnalyticPhase
from .transcriptions import GaussLobatto, Radau, ExplicitShooting, Analytic, \
    Birkhoff, PicardShooting
from .transcriptions.grid_data import GaussLobattoGrid, ChebyshevGaussLobattoGrid, \
    RadauGrid, UniformGrid, BirkhoffGrid
from .trajectory.trajectory import Trajectory
from .run_problem import run_problem
from .load_case import load_case
from ._options import options



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm


class _BrachistochroneTestODE(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input('v', val=np.zeros(nn), desc='velocity', units='m/s')

        self.add_input('g', val=9.80665 * np.ones(nn), desc='grav. acceleration', units='m/s/s')

        self.add_input('theta', val=np.zeros(nn), desc='angle of wire', units='rad')

        self.add_input('t_initial', val=-88.0, desc='start time of phase', units='s')

        self.add_input('t_duration', val=-89.0, desc='total duration of phase', units='s')

        self.add_input('time_phase', val=np.zeros(nn), desc='elapsed time of phase', units='s')

        self.add_input('time', val=np.zeros(nn), desc='time of phase', units='s')

        self.add_output('xdot', val=np.zeros(nn), desc='velocity component in x', units='m/s')

        self.add_output('ydot', val=np.zeros(nn), desc='velocity component in y', units='m/s')

        self.add_output('vdot', val=np.zeros(nn), desc='acceleration magnitude', units='m/s**2')

        self.add_output('check', val=np.zeros(nn), desc='check solution: v/sin(theta) = constant',
                        units='m/s')

        # Setup partials
        arange = np.arange(self.options['num_nodes'])

        self.declare_partials(of='vdot', wrt='g', rows=arange, cols=arange)
        self.declare_partials(of='vdot', wrt='theta', rows=arange, cols=arange)

        self.declare_partials(of='xdot', wrt='v', rows=arange, cols=arange)
        self.declare_partials(of='xdot', wrt='theta', rows=arange, cols=arange)

        self.declare_partials(of='ydot', wrt='v', rows=arange, cols=arange)
        self.declare_partials(of='ydot', wrt='theta', rows=arange, cols=arange)

        self.declare_partials(of='check', wrt='v', rows=arange, cols=arange)
        self.declare_partials(of='check', wrt='theta', rows=arange, cols=arange)

    def compute(self, inputs, outputs):
        theta = inputs['theta']
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)
        g = inputs['g']
        v = inputs['v']

        outputs['vdot'] = g * cos_theta
        outputs['xdot'] = v * sin_theta
        outputs['ydot'] = -v * cos_theta
        outputs['check'] = v / sin_theta

    def compute_partials(self, inputs, jacobian):
        theta = inputs['theta']
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)
        g = inputs['g']
        v = inputs['v']

        jacobian['vdot', 'g'] = cos_theta
        jacobian['vdot', 'theta'] = -g * sin_theta

        jacobian['xdot', 'v'] = sin_theta
        jacobian['xdot', 'theta'] = v * cos_theta

        jacobian['ydot', 'v'] = -cos_theta
        jacobian['ydot', 'theta'] = v * sin_theta

        jacobian['check', 'v'] = 1 / sin_theta
        jacobian['check', 'theta'] = -v * cos_theta / sin_theta**2


@use_tempdirs
class TestPhaseTimeTargets(unittest.TestCase):

    def _make_problem(self, transcription, num_seg, transcription_order=3, input_initial=False,
                      input_duration=False, time_name='time'):
        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()

        # Compute sparsity/coloring when run_driver is called
        p.driver.declare_coloring()

        t = {'gauss-lobatto': dm.GaussLobatto(num_segments=num_seg, order=transcription_order),
             'radau-ps': dm.Radau(num_segments=num_seg, order=transcription_order),
             'explicit-shooting': dm.ExplicitShooting(num_segments=num_seg, grid='radau-ps')}

        phase = dm.Phase(ode_class=_BrachistochroneTestODE, transcription=t[transcription])

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(initial_bounds=(1, 1), duration_bounds=(.5, 10), units='s',
                               time_phase_targets=['time_phase'], t_duration_targets=['t_duration'],
                               t_initial_targets=['t_initial'], targets=['time'],
                               input_initial=input_initial, input_duration=input_duration, name=time_name)

        phase.add_state('x', fix_initial=True, rate_source='xdot', units='m')
        phase.add_state('y', fix_initial=True, rate_source='ydot', units='m')
        phase.add_state('v', fix_initial=True, rate_source='vdot', targets=['v'], units='m/s')

        phase.add_control('theta', units='deg', rate_continuity=True, lower=0.01, upper=179.9, targets=['theta'])

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665, targets=['g'])

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)

        phase.timeseries_options['include_t_phase'] = True

        # Minimize time at the end of the phase
        phase.add_objective(time_name, loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()

        if input_duration:
            p.model.add_design_var('phase0.t_duration', lower=0, upper=3, scaler=1.0)

        p.setup(check=True, force_alloc_complex=True)

        phase.set_time_val(initial=0, duration=2.0)
        phase.set_state_val('x', (0, 10))
        phase.set_state_val('y', (10, 5))
        phase.set_state_val('v', (0, 9.9))
        phase.set_control_val('theta', (5, 100))
        phase.set_parameter_val('g', 9.80665)

        return p

    def test_gauss_lobatto(self):
        num_seg = 20

        for time_name in ('time', 'elapsed_time'):
            with self.subTest():

                p = self._make_problem('gauss-lobatto', num_seg, time_name=time_name)

                # Solve for the optimal trajectory
                p.run_driver()

                gd = p.model.phase0.options['transcription'].grid_data

                time_all = p[f'phase0.timeseries.{time_name}'].ravel()
                time_col = time_all[gd.subset_node_indices['col']]
                time_disc = time_all[gd.subset_node_indices['state_disc']]

                time_phase_all = p[f'phase0.timeseries.{time_name}_phase'].ravel()
                time_phase_col = time_phase_all[gd.subset_node_indices['col']]
                time_phase_disc = time_phase_all[gd.subset_node_indices['state_disc']]

                assert_near_equal(p['phase0.rhs_disc.time_phase'][-1], 1.8016, tolerance=1.0E-3)

                assert_near_equal(p['phase0.rhs_disc.t_initial'], p['phase0.t_initial'])
                assert_near_equal(p['phase0.rhs_col.t_initial'], p['phase0.t_initial'])

                assert_near_equal(p['phase0.rhs_disc.t_duration'], p['phase0.t_duration'])
                assert_near_equal(p['phase0.rhs_col.t_duration'], p['phase0.t_duration'])

                assert_near_equal(p['phase0.rhs_disc.time_phase'], time_phase_disc)
                assert_near_equal(p['phase0.rhs_col.time_phase'], time_phase_col)

                assert_near_equal(p['phase0.rhs_disc.time'], time_disc)
                assert_near_equal(p['phase0.rhs_col.time'], time_col)

                exp_out = p.model._get_subsystem('phase0').simulate()

                time_comp = exp_out.model.phase0._get_subsystem('time')
                integrator_comp = exp_out.model.phase0._get_subsystem('integrator')
                ode = exp_out.model.phase0._get_subsystem('ode')
                timeseries_comp = exp_out.model.phase0._get_subsystem('timeseries')

                time_comp_t_initial = time_comp.get_val('t_initial')
                integrator_comp_t_initial = integrator_comp.get_val('t_initial')
                ode_t_initial = ode.get_val('t_initial')

                time_comp_t_duration = time_comp.get_val('t_duration')
                integrator_comp_t_duration = integrator_comp.get_val('t_duration')
                ode_t_duration = ode.get_val('t_duration')
                ode_time_phase = ode.get_val('time_phase')
                timeseries_time_phase = timeseries_comp.get_val(f'{time_name}_phase')

                assert_near_equal(time_comp_t_initial, p['phase0.t_initial'])
                assert_near_equal(integrator_comp_t_initial, p['phase0.t_initial'])
                assert_near_equal(ode_t_initial, p['phase0.t_initial'])

                assert_near_equal(time_comp_t_duration, p['phase0.t_duration'])
                assert_near_equal(integrator_comp_t_duration, p['phase0.t_duration'])
                assert_near_equal(ode_t_duration, p['phase0.t_duration'])

                assert_near_equal(ode_time_phase.ravel(), timeseries_time_phase.ravel())

    def test_radau(self):
        for time_name in ('time', 'elapsed_time'):
            with self.subTest():
                num_seg = 20
                p = self._make_problem('radau-ps', num_seg, time_name=time_name)

                # Solve for the optimal trajectory
                p.run_driver()

                time_all = p[f'phase0.timeseries.{time_name}'].ravel()

                time_phase_all = p[f'phase0.timeseries.{time_name}_phase'].ravel()

                assert_near_equal(p['phase0.rhs_all.time_phase'][-1], 1.8016, tolerance=1.0E-3)

                assert_near_equal(p['phase0.rhs_all.t_initial'], p['phase0.t_initial'])

                assert_near_equal(p['phase0.rhs_all.t_duration'], p['phase0.t_duration'])

                assert_near_equal(p['phase0.rhs_all.time_phase'], time_phase_all)

                assert_near_equal(p['phase0.rhs_all.time'], time_all)

                exp_out = p.model.phase0.simulate()

                time_comp = exp_out.model.phase0._get_subsystem('time')
                integrator_comp = exp_out.model.phase0._get_subsystem('integrator')
                ode = exp_out.model.phase0._get_subsystem('ode')
                timeseries_comp = exp_out.model.phase0._get_subsystem('timeseries')

                time_comp_t_initial = time_comp.get_val('t_initial')
                integrator_comp_t_initial = integrator_comp.get_val('t_initial')
                ode_t_initial = ode.get_val('t_initial')

                time_comp_t_duration = time_comp.get_val('t_duration')
                integrator_comp_t_duration = integrator_comp.get_val('t_duration')
                ode_t_duration = ode.get_val('t_duration')
                ode_time_phase = ode.get_val('time_phase')
                timeseries_time_phase = timeseries_comp.get_val(f'{time_name}_phase')

                assert_near_equal(time_comp_t_initial, p['phase0.t_initial'])
                assert_near_equal(integrator_comp_t_initial, p['phase0.t_initial'])
                assert_near_equal(ode_t_initial, p['phase0.t_initial'])

                assert_near_equal(time_comp_t_duration, p['phase0.t_duration'])
                assert_near_equal(integrator_comp_t_duration, p['phase0.t_duration'])
                assert_near_equal(ode_t_duration, p['phase0.t_duration'])

                assert_near_equal(ode_time_phase.ravel(), timeseries_time_phase.ravel())

    def test_explicit_shooting(self):
        num_seg = 5

        for time_name in ('time', 'elapsed_time'):
            with self.subTest():
                p = self._make_problem('explicit-shooting', num_seg, time_name=time_name)

                # Solve for the optimal trajectory
                p.run_driver()

                time_all = p[f'phase0.timeseries.{time_name}']

                time_phase_all = p[f'phase0.timeseries.{time_name}_phase']

                assert_near_equal(p['phase0.t_phase'][-1], 1.8016, tolerance=1.0E-3)

                assert_near_equal(p['phase0.t_initial'], p['phase0.integrator.t_initial'])

                assert_near_equal(p['phase0.t_duration'], p['phase0.integrator.t_duration'])

                assert_near_equal(np.atleast_2d(p['phase0.t_phase']).T, time_phase_all)

                assert_near_equal(np.atleast_2d(p['phase0.t']).T, time_all)

    def test_gauss_lobatto_targets_are_inputs(self):
        num_seg = 20
        p = self._make_problem('gauss-lobatto', num_seg, input_initial=True, input_duration=True)

        # Solve for the optimal trajectory
        p.run_driver()

        gd = p.model.phase0.options['transcription'].grid_data

        time_all = p['phase0.t']
        time_col = time_all[gd.subset_node_indices['col']]
        time_disc = time_all[gd.subset_node_indices['state_disc']]
        time_segends = np.reshape(time_all[gd.subset_node_indices['segment_ends']],  # noqa: F841
                                  (gd.num_segments, 2))

        time_phase_all = p['phase0.t_phase']
        time_phase_col = time_phase_all[gd.subset_node_indices['col']]
        time_phase_disc = time_phase_all[gd.subset_node_indices['state_disc']]
        time_phase_segends = np.reshape(time_phase_all[gd.subset_node_indices['segment_ends']],  # noqa: F841
                                        (gd.num_segments, 2))

        assert_near_equal(p['phase0.rhs_disc.time_phase'][-1], 1.8016, tolerance=1.0E-3)

        assert_near_equal(p['phase0.rhs_disc.t_initial'], p['phase0.t_initial'])
        assert_near_equal(p['phase0.rhs_col.t_initial'], p['phase0.t_initial'])

        assert_near_equal(p['phase0.rhs_disc.t_duration'], p['phase0.t_duration'])
        assert_near_equal(p['phase0.rhs_col.t_duration'], p['phase0.t_duration'])

        assert_near_equal(p['phase0.rhs_disc.time_phase'], time_phase_disc)
        assert_near_equal(p['phase0.rhs_col.time_phase'], time_phase_col)

        assert_near_equal(p['phase0.rhs_disc.time'], time_disc)
        assert_near_equal(p['phase0.rhs_col.time'], time_col)

        exp_out = p.model._get_subsystem('phase0').simulate()

        time_comp = exp_out.model.phase0._get_subsystem('time')
        integrator_comp = exp_out.model.phase0._get_subsystem('integrator')
        ode = exp_out.model.phase0._get_subsystem('ode')
        timeseries_comp = exp_out.model.phase0._get_subsystem('timeseries')

        time_comp_t_initial = time_comp.get_val('t_initial')
        integrator_comp_t_initial = integrator_comp.get_val('t_initial')
        ode_t_initial = ode.get_val('t_initial')

        time_comp_t_duration = time_comp.get_val('t_duration')
        integrator_comp_t_duration = integrator_comp.get_val('t_duration')
        ode_t_duration = ode.get_val('t_duration')
        ode_time_phase = ode.get_val('time_phase')
        timeseries_time_phase = timeseries_comp.get_val('time_phase')

        assert_near_equal(time_comp_t_initial, p['phase0.t_initial'])
        assert_near_equal(integrator_comp_t_initial, p['phase0.t_initial'])
        assert_near_equal(ode_t_initial, p['phase0.t_initial'])

        assert_near_equal(time_comp_t_duration, p['phase0.t_duration'])
        assert_near_equal(integrator_comp_t_duration, p['phase0.t_duration'])
        assert_near_equal(ode_t_duration, p['phase0.t_duration'])

        assert_near_equal(ode_time_phase.ravel(), timeseries_time_phase.ravel())

    def test_radau_targets_are_inputs(self):
        num_seg = 20
        p = self._make_problem('radau-ps', num_seg, input_initial=True, input_duration=True)

        # Solve for the optimal trajectory
        p.run_driver()

        time_all = p['phase0.t']

        time_phase_all = p['phase0.t_phase']

        assert_near_equal(p['phase0.rhs_all.time_phase'][-1], 1.8016, tolerance=1.0E-3)

        assert_near_equal(p['phase0.rhs_all.t_initial'], p['phase0.t_initial'])

        assert_near_equal(p['phase0.rhs_all.t_duration'], p['phase0.t_duration'])

        assert_near_equal(p['phase0.rhs_all.time_phase'], time_phase_all)

        assert_near_equal(p['phase0.rhs_all.time'], time_all)

        exp_out = p.model._get_subsystem('phase0').simulate()

        time_comp = exp_out.model.phase0._get_subsystem('time')
        integrator_comp = exp_out.model.phase0._get_subsystem('integrator')
        ode = exp_out.model.phase0._get_subsystem('ode')
        timeseries_comp = exp_out.model.phase0._get_subsystem('timeseries')

        time_comp_t_initial = time_comp.get_val('t_initial')
        integrator_comp_t_initial = integrator_comp.get_val('t_initial')
        ode_t_initial = ode.get_val('t_initial')

        time_comp_t_duration = time_comp.get_val('t_duration')
        integrator_comp_t_duration = integrator_comp.get_val('t_duration')
        ode_t_duration = ode.get_val('t_duration')
        ode_time_phase = ode.get_val('time_phase')
        timeseries_time_phase = timeseries_comp.get_val('time_phase')

        assert_near_equal(time_comp_t_initial, p['phase0.t_initial'])
        assert_near_equal(integrator_comp_t_initial, p['phase0.t_initial'])
        assert_near_equal(ode_t_initial, p['phase0.t_initial'])

        assert_near_equal(time_comp_t_duration, p['phase0.t_duration'])
        assert_near_equal(integrator_comp_t_duration, p['phase0.t_duration'])
        assert_near_equal(ode_t_duration, p['phase0.t_duration'])

        assert_near_equal(ode_time_phase.ravel(), timeseries_time_phase.ravel())


if __name__ == "__main__":
    unittest.main()



import numpy as np
import openmdao.api as om


class MoonLandingProblemODE(om.ExplicitComponent):
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        # inputs
        self.add_input('h', val=np.ones(nn), units=None, desc='Altitude')
        self.add_input('v', val=np.ones(nn), units='1/s', desc='Velocity')
        self.add_input('m', val=np.ones(nn), units=None, desc='Mass')
        self.add_input('T', val=np.ones(nn), units=None, desc='Thrust')

        # outputs
        self.add_output('h_dot', val=np.ones(nn), units='1/s', desc='Rate of change of Altitude')
        self.add_output('v_dot', val=np.ones(nn), units='1/s**2', desc='Rate of change of Velocity')
        self.add_output('m_dot', val=np.ones(nn), units='1/s', desc='Rate of change of Mass')

        # partials
        ar = np.arange(nn)
        self.declare_partials(of='h_dot', wrt='v', rows=ar, cols=ar, val=1.0)
        self.declare_partials(of='v_dot', wrt='m', rows=ar, cols=ar)
        self.declare_partials(of='v_dot', wrt='T', rows=ar, cols=ar)
        self.declare_partials(of='m_dot', wrt='T', rows=ar, cols=ar, val=-1/2.349)
        self.declare_partials(of='m_dot', wrt='T', rows=ar, cols=ar, val=-1/2.349)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        v = inputs['v']
        m = inputs['m']
        T = inputs['T']

        outputs['h_dot'] = v
        outputs['v_dot'] = -1 + T/m
        outputs['m_dot'] = -T/2.349

    def compute_partials(self, inputs, partials, discrete_inputs=None):
        m = inputs['m']
        T = inputs['T']

        partials['v_dot', 'T'] = 1/m
        partials['v_dot', 'm'] = -T/m**2



import numpy as np
import openmdao.api as om

from .crm_data import h_bp, alpha_bp, mach_bp, eta_bp, CL_data, CD_data, CM_data


class AeroCoefComp(om.MetaModelStructuredComp):
    """ Interpolates aerodynamic coefficients for the NASA Common Research Model. """

    def setup(self):
        nn = self.options['vec_size']
        self.add_input(name='mach', val=0.2 * np.ones(nn), units=None, training_data=mach_bp)
        self.add_input(name='alpha', val=0.0 * np.ones(nn), units='deg', training_data=alpha_bp)
        self.add_input(name='alt', val=0.0 * np.ones(nn), units='ft', training_data=h_bp)
        self.add_input(name='eta', val=0.0 * np.ones(nn), units='deg', training_data=eta_bp)

        self.add_output(name='CL', val=np.zeros(nn), units=None, training_data=CL_data)
        self.add_output(name='CD', val=np.zeros(nn), units=None, training_data=CD_data + 0.015)
        self.add_output(name='CM', val=np.zeros(nn), units=None, training_data=CM_data)



"""Test Dymos Linkage report GUI with using Playwright."""
import unittest
import os
try:
    import playwright
except ImportError:
    playwright = None


if playwright is not None:
    os.system("playwright install")
    from dymos.visualization.linkage.test.linkage_report_ui_test import dymos_linkage_gui_test_case  # noqa: E402, F401


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse
from openmdao.utils.assert_utils import assert_near_equal
import dymos as dm
from dymos.utils.misc import om_version


@use_tempdirs
class TestBrachExecCompODE(unittest.TestCase):

    @require_pyoptsparse(optimizer='SLSQP')
    def _make_problem(self, transcription='gauss-lobatto', num_segments=8, transcription_order=3,
                      compressed=True, optimizer='SLSQP', run_driver=True, force_alloc_complex=False,
                      solve_segments=False):

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = optimizer
        p.driver.declare_coloring(tol=1.0E-12)

        if transcription == 'gauss-lobatto':
            t = dm.GaussLobatto(num_segments=num_segments,
                                order=transcription_order,
                                compressed=compressed)
        elif transcription == 'radau-ps':
            t = dm.Radau(num_segments=num_segments,
                         order=transcription_order,
                         compressed=compressed)
        elif transcription == 'birkhoff':
            t = dm.Birkhoff(num_nodes=transcription_order+1,
                            grid_type='cgl')
        elif transcription.startswith('picard'):
            grid_type = transcription.split('-')[-1]
            t = dm.PicardShooting(nodes_per_seg=5, num_segments=8, grid_type=grid_type)

        def ode(num_nodes):
            return om.ExecComp(['vdot = g * cos(theta)',
                                'xdot = v * sin(theta)',
                                'ydot = -v * cos(theta)'],
                               g={'val': 9.80665, 'units': 'm/s**2'},
                               v={'shape': (num_nodes,), 'units': 'm/s'},
                               theta={'shape': (num_nodes,), 'units': 'rad'},
                               vdot={'shape': (num_nodes,),
                                     'units': 'm/s**2',
                                     'tags': ['dymos.state_rate_source:v']},
                               xdot={'shape': (num_nodes,),
                                     'units': 'm/s',
                                     'tags': ['dymos.state_rate_source:x']},
                               ydot={'shape': (num_nodes,),
                                     'units': 'm/s',
                                     'tags': ['dymos.state_rate_source:y']},
                               has_diag_partials=True)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=ode, transcription=t)
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False, solve_segments=solve_segments)
        phase.add_state('y', fix_initial=True, fix_final=False, solve_segments=solve_segments)

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=True, fix_final=False, solve_segments=solve_segments)

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', static_target=True)

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.setup(check=['unconnected_inputs'], force_alloc_complex=force_alloc_complex)
        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])

        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)

        dm.run_problem(p, run_driver=run_driver, simulate=True)

        return p

    def run_asserts(self, c):
        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = c.get_outputs_dir() / sol_db
            sim_db = c.model.traj0.sim_prob.get_outputs_dir() / sim_db

        for db in [sol_db, sim_db]:
            c = om.CaseReader(db).get_case('final')

            t_initial = c.get_val('traj0.phase0.timeseries.time')[0]
            tf = c.get_val('traj0.phase0.timeseries.time')[-1]

            x0 = c.get_val('traj0.phase0.timeseries.x')[0]
            xf = c.get_val('traj0.phase0.timeseries.x')[-1]

            y0 = c.get_val('traj0.phase0.timeseries.y')[0]
            yf = c.get_val('traj0.phase0.timeseries.y')[-1]

            v0 = c.get_val('traj0.phase0.timeseries.v')[0]
            vf = c.get_val('traj0.phase0.timeseries.v')[-1]

            g = c.get_val('traj0.phase0.parameter_vals:g')[0]

            thetaf = c.get_val('traj0.phase0.timeseries.theta')[-1]

            assert_near_equal(t_initial, 0.0)
            assert_near_equal(x0, 0.0)
            assert_near_equal(y0, 10.0)
            assert_near_equal(v0, 0.0)

            assert_near_equal(tf, 1.8016, tolerance=0.01)
            assert_near_equal(xf, 10.0, tolerance=0.01)
            assert_near_equal(yf, 5.0, tolerance=0.01)
            assert_near_equal(vf, 9.902, tolerance=0.01)
            assert_near_equal(g, 9.80665, tolerance=0.01)

            assert_near_equal(thetaf, 100.12, tolerance=0.01)

    def test_ex_brachistochrone_radau_uncompressed(self):
        p = self._make_problem(transcription='radau-ps', compressed=False)
        self.run_asserts(p)

    def test_ex_brachistochrone_gl_uncompressed(self):
        p = self._make_problem(transcription='gauss-lobatto', compressed=False)
        self.run_asserts(p)

    def test_ex_brachistochrone_birkhoff_uncompressed(self):
        p = self._make_problem(transcription='birkhoff', transcription_order=10)
        self.run_asserts(p)

    def test_ex_brachistochrone_picard_shooting(self):
        p = self._make_problem(transcription='picard-cgl', run_driver=True)
        self.run_asserts(p)


@use_tempdirs
class TestInvalidCallableODEClass(unittest.TestCase):
    @require_pyoptsparse(optimizer='SLSQP')
    def test_invalid_callable(self):
        num_segments = 10
        transcription_order = 3
        compressed = False

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring(tol=1.0E-12)

        t = dm.Radau(num_segments=num_segments,
                     order=transcription_order,
                     compressed=compressed)

        def ode(num_nodes):
            return num_nodes*2

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=ode, transcription=t)
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False, rate_source='xdot')
        phase.add_state('y', fix_initial=True, fix_final=False, rate_source='ydot')

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=True, fix_final=False, rate_source='vdot')

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', static_target=True)

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        with self.assertRaises(expected_exception=ValueError) as e:
            p.setup()
        expected = "When provided as a callable, ode_class must return an instance of " \
                   "openmdao.core.System.  Got <class 'int'>"
        self.assertEqual(expected, str(e.exception))


class CallableBrachistochroneODE(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)
        self.matrix_free = False

    def __call__(self, num_nodes, **kwargs):
        from copy import deepcopy
        ret = deepcopy(self)
        ret.options['num_nodes'] = num_nodes
        return ret

    def setup(self):
        nn = self.options['num_nodes']

        # Inputs
        self.add_input('v', val=np.zeros(nn), desc='velocity', units='m/s')

        self.add_input('g', val=9.80665, desc='grav. acceleration', units='m/s/s')

        self.add_input('theta', val=np.ones(nn), desc='angle of wire', units='rad')

        self.add_output('xdot', val=np.zeros(nn), desc='velocity component in x', units='m/s',
                        tags=['dymos.state_rate_source:x', 'dymos.state_units:m'])

        self.add_output('ydot', val=np.zeros(nn), desc='velocity component in y', units='m/s',
                        tags=['dymos.state_rate_source:y', 'dymos.state_units:m'])

        self.add_output('vdot', val=np.zeros(nn), desc='acceleration magnitude', units='m/s**2',
                        tags=['dymos.state_rate_source:v', 'dymos.state_units:m/s'])

        self.declare_partials(of='*', wrt='*', method='cs')

    def compute(self, inputs, outputs):
        theta = inputs['theta']
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)
        g = inputs['g']
        v = inputs['v']

        outputs['vdot'] = g * cos_theta
        outputs['xdot'] = v * sin_theta
        outputs['ydot'] = -v * cos_theta


@use_tempdirs
class TestBrachCallableODE(unittest.TestCase):

    def setUp(self):
        self.ode = CallableBrachistochroneODE(num_nodes=1)

    @require_pyoptsparse(optimizer='SLSQP')
    def _make_problem(self, transcription='gauss-lobatto', num_segments=8, transcription_order=3,
                      compressed=True, optimizer='SLSQP', run_driver=True,
                      force_alloc_complex=False,
                      solve_segments=False):

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = optimizer
        p.driver.declare_coloring(tol=1.0E-12)

        if transcription == 'gauss-lobatto':
            t = dm.GaussLobatto(num_segments=num_segments,
                                order=transcription_order,
                                compressed=compressed)
        elif transcription == 'radau-ps':
            t = dm.Radau(num_segments=num_segments,
                         order=transcription_order,
                         compressed=compressed)

        traj = dm.Trajectory()
        phase = dm.Phase(ode_class=self.ode, transcription=t)
        p.model.add_subsystem('traj0', traj)
        traj.add_phase('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', fix_initial=True, fix_final=False, solve_segments=solve_segments,
                        rate_source='xdot')
        phase.add_state('y', fix_initial=True, fix_final=False, solve_segments=solve_segments,
                        rate_source='ydot')

        # Note that by omitting the targets here Dymos will automatically attempt to connect
        # to a top-level input named 'v' in the ODE, and connect to nothing if it's not found.
        phase.add_state('v', fix_initial=True, fix_final=False, solve_segments=solve_segments,
                        rate_source='vdot')

        phase.add_control('theta',
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', static_target=True)

        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)
        # Minimize time at the end of the phase
        phase.add_objective('time_phase', loc='final', scaler=10)

        p.setup(check=['unconnected_inputs'], force_alloc_complex=force_alloc_complex)

        phase.set_time_val(initial=0.0, duration=2.0)

        phase.set_state_val('x', [0, 10])
        phase.set_state_val('y', [10, 5])
        phase.set_state_val('v', [0, 9.9])

        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)

        dm.run_problem(p, run_driver=run_driver, simulate=True)

        return p

    def run_asserts(self, c):
        sol_db = 'dymos_solution.db'
        sim_db = 'dymos_simulation.db'
        if om_version()[0] > (3, 34, 2):
            sol_db = c.get_outputs_dir() / sol_db
            sim_db = c.model.traj0.sim_prob.get_outputs_dir() / sim_db

        for db in (sol_db, sim_db):
            c = om.CaseReader(db).get_case('final')

            t_initial = c.get_val('traj0.phase0.timeseries.time')[0]
            tf = c.get_val('traj0.phase0.timeseries.time')[-1]

            x0 = c.get_val('traj0.phase0.timeseries.x')[0]
            xf = c.get_val('traj0.phase0.timeseries.x')[-1]

            y0 = c.get_val('traj0.phase0.timeseries.y')[0]
            yf = c.get_val('traj0.phase0.timeseries.y')[-1]

            v0 = c.get_val('traj0.phase0.timeseries.v')[0]
            vf = c.get_val('traj0.phase0.timeseries.v')[-1]

            g = c.get_val('traj0.phase0.parameter_vals:g')[0]

            thetaf = c.get_val('traj0.phase0.timeseries.theta')[-1]

            assert_near_equal(t_initial, 0.0)
            assert_near_equal(x0, 0.0)
            assert_near_equal(y0, 10.0)
            assert_near_equal(v0, 0.0)

            assert_near_equal(tf, 1.8016, tolerance=0.01)
            assert_near_equal(xf, 10.0, tolerance=0.01)
            assert_near_equal(yf, 5.0, tolerance=0.01)
            assert_near_equal(vf, 9.902, tolerance=0.01)
            assert_near_equal(g, 9.80665, tolerance=0.01)

            assert_near_equal(thetaf, 100.12, tolerance=0.01)

    def test_ex_brachistochrone_radau_uncompressed(self):
        p = self._make_problem(transcription='radau-ps', compressed=False)
        self.run_asserts(p)

    def test_in_series(self):
        p = self._make_problem(transcription='gauss-lobatto', compressed=False)
        self.run_asserts(p)

        p = self._make_problem(transcription='radau-ps', compressed=False)
        self.run_asserts(p)



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos.examples.cart_pole.cartpole_dynamics import CartPoleDynamics


@use_tempdirs
class TestCartPoleOptimization(unittest.TestCase):
    @require_pyoptsparse(optimizer="SNOPT")
    def test_optimization(self):

        p = om.Problem()

        # --- instantiate trajectory and phase, setup transcription ---
        traj = dm.Trajectory()
        p.model.add_subsystem("traj", traj)
        phase = dm.Phase(
            transcription=dm.GaussLobatto(num_segments=40, order=3, compressed=True, solve_segments=False),
            ode_class=CartPoleDynamics,
        )
        # NOTE: set solve_segments=True to do solver-based shooting
        traj.add_phase("phase", phase)

        # --- set state and control variables ---
        phase.set_time_options(fix_initial=True, fix_duration=True, duration_val=2.0, units="s")
        # declare state variables. You can also set lower/upper bounds and scalings here.
        phase.add_state("x", fix_initial=True, lower=-2, upper=2, rate_source="x_dot", shape=(1,), ref=1, defect_ref=1, units="m")
        phase.add_state("x_dot", fix_initial=True, rate_source="x_dotdot", shape=(1,), ref=1, defect_ref=1, units="m/s")
        phase.add_state("theta", fix_initial=True, rate_source="theta_dot", shape=(1,), ref=1, defect_ref=1, units="rad")
        phase.add_state("theta_dot", fix_initial=True, rate_source="theta_dotdot", shape=(1,), ref=1, defect_ref=1, units="rad/s")
        phase.add_state(
            "energy", fix_initial=True, rate_source="e_dot", shape=(1,), ref=1, defect_ref=1, units="N**2*s"
        )  # integration of force**2. This does not have the energy unit, but I call it "energy" anyway.

        # declare control inputs
        phase.add_control("f", fix_initial=False, rate_continuity=False, lower=-20, upper=20, shape=(1,), ref=0.01, units="N")

        # add cart-pole parameters (set static_target=True because these params are not time-depencent)
        phase.add_parameter("m_cart", val=1.0, units="kg", static_target=True)
        phase.add_parameter("m_pole", val=0.3, units="kg", static_target=True)
        phase.add_parameter("l_pole", val=0.5, units="m", static_target=True)

        # --- set terminal constraint ---
        # alternatively, you can impose those by setting `fix_final=True` in phase.add_state()
        phase.add_boundary_constraint("x", loc="final", equals=1, ref=1.0, units="m")  # final horizontal displacement
        phase.add_boundary_constraint("theta", loc="final", equals=np.pi, ref=1.0, units="rad")  # final pole angle
        phase.add_boundary_constraint("x_dot", loc="final", equals=0, ref=1.0, units="m/s")  # 0 velocity at the and
        phase.add_boundary_constraint("theta_dot", loc="final", equals=0, ref=1.0, units="rad/s")  # 0 angular velocity at the end
        phase.add_boundary_constraint("f", loc="final", equals=0, ref=1.0, units="N")  # 0 force at the end

        # --- set objective function ---
        # we minimize the integral of force**2.
        phase.add_objective("energy", loc="final", ref=1.0)

        # --- configure optimizer ---
        p.driver = om.pyOptSparseDriver()
        p.driver.options["optimizer"] = "IPOPT"
        # IPOPT options
        p.driver.opt_settings['mu_init'] = 1e-1
        p.driver.opt_settings['max_iter'] = 600
        p.driver.opt_settings['constr_viol_tol'] = 1e-6
        p.driver.opt_settings['compl_inf_tol'] = 1e-6
        p.driver.opt_settings['tol'] = 1e-5
        p.driver.opt_settings['print_level'] = 0
        p.driver.opt_settings['nlp_scaling_method'] = 'gradient-based'
        p.driver.opt_settings['alpha_for_y'] = 'safer-min-dual-infeas'
        p.driver.opt_settings['mu_strategy'] = 'monotone'
        p.driver.opt_settings['bound_mult_init_method'] = 'mu-based'
        p.driver.options['print_results'] = False

        # declare total derivative coloring to accelerate the UDE linear solves
        p.driver.declare_coloring()

        p.setup(check=False)

        # --- set initial guess ---
        # The initial condition of cart-pole (i.e., state values at time 0) is set here
        # because we set `fix_initial=True` when declaring the states.
        phase.set_time_val(initial=0.0)  # set initial time to 0.
        phase.set_state_val('x', vals=[0, 1, 1], time_vals=[0, 1, 2], units='m')
        phase.set_state_val('x_dot', vals=[0, 0.1, 0], time_vals=[0, 1, 2], units='m/s')
        phase.set_state_val('theta', vals=[0, np.pi/2, np.pi], time_vals=[0, 1, 2], units='rad')
        phase.set_state_val('theta_dot', vals=[0, 1, 0], time_vals=[0, 1, 2], units='rad/s')
        phase.set_state_val('energy', vals=[0, 30, 60], time_vals=[0, 1, 2])
        phase.set_control_val('f', vals=[3, -1, 0], time_vals=[0, 1, 2], units='N')

        # --- run optimization ---
        dm.run_problem(p, run_driver=True, simulate=False, simulate_kwargs={"method": "Radau", "times_per_seg": 10})

        # --- check outputs ---
        # objective value
        obj = p.get_val("traj.phase.states:energy", units="N**2*s")[-1]
        assert_near_equal(obj, 58.8839489745, tolerance=1e-3)


if __name__ == "___main__":
    unittest.main()



from __future__ import print_function, division, absolute_import

import os
import unittest
import warnings

import numpy as np
import openmdao.api as om

from openmdao.utils.assert_utils import assert_near_equal, assert_check_partials
from openmdao.utils.general_utils import printoptions
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos.examples.hyper_sensitive.hyper_sensitive_ode import HyperSensitiveODE


@use_tempdirs
class TestHyperSensitive(unittest.TestCase):

    def tearDown(self):
        for filename in ['total_coloring.pkl', 'SLSQP.out', 'SNOPT_print.out']:
            if os.path.exists(filename):
                os.remove(filename)

    @require_pyoptsparse(optimizer='SLSQP')
    def make_problem(self, transcription=dm.GaussLobatto, optimizer='SLSQP', numseg=30, order=3,
                     solve_segments=False, tf=10):

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()
        p.driver.options['optimizer'] = optimizer

        if optimizer == 'SNOPT':
            p.driver.declare_coloring()
            p.driver.opt_settings['iSumm'] = 6
            p.driver.opt_settings['Major iterations limit'] = 500
            p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
            p.driver.opt_settings['Major optimality tolerance'] = 1.0E-6
        elif optimizer == 'IPOPT':
            p.driver.opt_settings['print_level'] = 0
            p.driver.opt_settings['mu_strategy'] = 'adaptive'
            p.driver.opt_settings['bound_mult_init_method'] = 'mu-based'
            p.driver.opt_settings['mu_init'] = 0.01
            p.driver.opt_settings['alpha_for_y'] = 'safer-min-dual-infeas'
            p.driver.opt_settings['nlp_scaling_method'] = 'gradient-based'
            p.driver.opt_settings['tol'] = 1.0E-6
            p.driver.declare_coloring()
        else:
            p.driver.declare_coloring()

        if transcription == 'gauss-lobatto':
            t = dm.GaussLobatto(num_segments=numseg, order=3)
        elif transcription == 'radau-ps':
            t = dm.Radau(num_segments=numseg, order=3)
        elif transcription == 'birkhoff':
            t = dm.Birkhoff(num_nodes=order + 1, grid_type='lgl',
                            solve_segments=solve_segments)

        traj = p.model.add_subsystem('traj', dm.Trajectory())
        phase0 = traj.add_phase('phase0', dm.Phase(ode_class=HyperSensitiveODE, transcription=t))
        phase0.set_time_options(fix_initial=True, fix_duration=True)
        phase0.add_state('x', fix_initial=False, fix_final=False, rate_source='x_dot', ref=100)
        phase0.add_state('xL', fix_initial=True, fix_final=False, rate_source='L', ref=100)
        phase0.add_control('u', opt=True, targets=['u'])

        phase0.add_boundary_constraint('x', loc='initial', equals=1.5)
        phase0.add_boundary_constraint('x', loc='final', equals=1)

        phase0.add_objective('xL', loc='final')

        phase0.set_refine_options(refine=True, tol=1e-7, max_order=14)
        phase0.set_simulate_options(times_per_seg=100)

        p.set_solver_print(0)
        p.setup(check=True, force_alloc_complex=True)

        phase0.set_time_val(initial=0, duration=tf)
        phase0.set_state_val('x', [1.5, 1])
        phase0.set_state_val('xL', [0, 1])
        phase0.set_control_val('u', [-0.6, 2.4])

        return p

    @staticmethod
    def solution(tf):
        sqrt_two = np.sqrt(2)
        val = sqrt_two * tf
        c1 = (1.5 * np.exp(-val) - 1) / (np.exp(-val) - np.exp(val))
        c2 = (1 - 1.5 * np.exp(val)) / (np.exp(-val) - np.exp(val))

        ui = c1 * (1 + sqrt_two) + c2 * (1 - sqrt_two)
        uf = c1 * (1 + sqrt_two) * np.exp(val) + c2 * (1 - sqrt_two) * np.exp(-val)
        J = 0.5 * (c1 ** 2 * (1 + sqrt_two) * np.exp(2 * val) + c2 ** 2 * (1 - sqrt_two) * np.exp(-2 * val) -
                   (1 + sqrt_two) * c1 ** 2 - (1 - sqrt_two) * c2 ** 2)
        return ui, uf, J

    def test_partials(self):
        p = self.make_problem(transcription='radau-ps', optimizer='SLSQP')
        p.run_model()
        with printoptions(linewidth=1024, edgeitems=100):
            cpd = p.check_partials(method='cs', compact_print=True, out_stream=None)
            assert_check_partials(cpd)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_hyper_sensitive_radau(self):
        tf = 10
        p = self.make_problem(transcription='radau-ps', optimizer='IPOPT', tf=tf)
        dm.run_problem(p, refine_iteration_limit=5)
        ui, uf, J = self.solution(tf)

        assert_near_equal(p.get_val('traj.phase0.timeseries.u')[0],
                          ui,
                          tolerance=5e-6)

        assert_near_equal(p.get_val('traj.phase0.timeseries.u')[-1],
                          uf,
                          tolerance=5e-6)

        assert_near_equal(p.get_val('traj.phase0.timeseries.xL')[-1],
                          J,
                          tolerance=1e-6)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_hyper_sensitive_birkhoff(self):
        tf = 10
        p = self.make_problem(transcription='birkhoff', optimizer='IPOPT', order=51, tf=tf)

        dm.run_problem(p, simulate=True, make_plots=True)
        ui, uf, J = self.solution(tf=tf)

        assert_near_equal(p.get_val('traj.phase0.timeseries.u')[0],
                          ui,
                          tolerance=1e-3)

        assert_near_equal(p.get_val('traj.phase0.timeseries.u')[-1],
                          uf,
                          tolerance=1e-3)

        assert_near_equal(p.get_val('traj.phase0.timeseries.xL')[-1],
                          J,
                          tolerance=1e-6)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_hyper_sensitive_gauss_lobatto(self):
        tf = 10
        p = self.make_problem(transcription='gauss-lobatto', optimizer='IPOPT', tf=tf)
        dm.run_problem(p, refine_iteration_limit=5,)

        ui, uf, J = self.solution(tf=tf)

        assert_near_equal(p.get_val('traj.phase0.timeseries.u')[0],
                          ui,
                          tolerance=1e-4)

        assert_near_equal(p.get_val('traj.phase0.timeseries.u')[-1],
                          uf,
                          tolerance=1e-4)

        assert_near_equal(p.get_val('traj.phase0.timeseries.xL')[-1],
                          J,
                          tolerance=1e-4)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_refinement_warning(self):
        p = self.make_problem(transcription='radau-ps', optimizer='IPOPT')

        msg = "Refinement not performed. Set run_driver to True to perform refinement."

        with warnings.catch_warnings(record=True) as ctx:
            warnings.simplefilter('always')
            dm.run_problem(p, run_driver=False, refine_iteration_limit=10)

        self.assertIn(msg, [str(w.message) for w in ctx])



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm


n_traj = 4


class MyComp(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('n_traj', types=int)
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']
        n_traj = self.options['n_traj']
        self.add_input('time', val=np.zeros(nn), units='s')
        self.add_input('alpha', shape=np.zeros((n_traj, 2)).shape, units='m')
        self.add_output('y', val=np.zeros(nn), units='1/s')

    def compute(self, inputs, outputs):
        pass

    def compute_partials(self, inputs, partials):
        pass


class MyODE(om.Group):

    def initialize(self):
        self.options.declare('num_nodes', types=int)
        self.options.declare('n_traj', default=2, types=int)

    def setup(self):
        nn = self.options['num_nodes']
        n_traj = self.options['n_traj']

        self.add_subsystem(name='comp',
                           subsys=MyComp(num_nodes=nn,
                                         n_traj=n_traj))


@use_tempdirs
class TestStaticParameters(unittest.TestCase):

    def test_radau(self):

        p = om.Problem(model=om.Group())

        phase = dm.Phase(ode_class=MyODE,
                         ode_init_kwargs={'n_traj': n_traj},
                         transcription=dm.Radau(num_segments=25,
                                                order=3,
                                                compressed=True))

        phase.set_time_options(units='s', targets=['comp.time'])
        phase.add_state(name='F', rate_source='comp.y')
        phase.add_parameter('alpha', val=np.ones((n_traj, 2)), units='m',
                            targets='comp.alpha', static_target=True)

        p.model.add_subsystem('phase0', phase)

        try:
            p.setup()
        except Exception as e:
            self.fail('Exception encountered in setup:\n' + str(e))

    def test_gauss_lobatto(self):

        p = om.Problem(model=om.Group())

        phase = dm.Phase(ode_class=MyODE,
                         ode_init_kwargs={'n_traj': n_traj},
                         transcription=dm.GaussLobatto(num_segments=25,
                                                       order=3,
                                                       compressed=True))

        phase.set_time_options(units='s', targets=['comp.time'])
        phase.add_state(name='F', rate_source='comp.y')
        phase.add_parameter('alpha', val=np.ones((n_traj, 2)), units='m',
                            targets='comp.alpha', static_target=True)

        p.model.add_subsystem('phase0', phase)

        try:
            p.setup()
        except Exception as e:
            self.fail('Exception encountered in setup:\n' + str(e))

    def test_static_and_dynamic_error(self):

        phase = dm.Phase(ode_class=MyODE,
                         ode_init_kwargs={'n_traj': n_traj},
                         transcription=dm.Radau(num_segments=25,
                                                order=3,
                                                compressed=True))

        with self.assertRaises(ValueError) as e:
            phase.add_parameter('alpha', val=np.ones((n_traj, 2)), units='m',
                                targets='comp.alpha', dynamic=False, static_target=True)

        expected_msg = "Both the deprecated 'dynamic' option and option 'static_target' were\n" \
                       "specified for parameter 'alpha'. Going forward, please use only\n" \
                       "option static_targets. Options 'dynamic' and 'static_target'\n" \
                       "will be removed in Dymos 2.0.0."

        self.assertEqual(expected_msg, str(e.exception))

    def test_static_and_dynamic_error_in_traj(self):

        t = dm.Trajectory()

        with self.assertRaises(ValueError) as e:
            t.add_parameter('alpha', val=np.ones((n_traj, 2)), units='m',
                            targets={'p': ['comp.alpha']}, dynamic=False, static_target=True)

        expected_msg = "Both the deprecated 'dynamic' option and option 'static_target' were\n" \
                       "specified for parameter 'alpha'. Going forward, please use only\n" \
                       "option static_targets. Options 'dynamic' and 'static_target'\n" \
                       "will be removed in Dymos 2.0.0."

        self.assertEqual(expected_msg, str(e.exception))


if __name__ == '__main__':  # pragma: no cover
    unittest.main()
